import os.path
from logging import getLogger
from typing import Optional

from colorama import Fore, Style
import mne
from numpy import nanmin
from numpy.typing import NDArray
from pandas import DataFrame, Index
from pathlib import Path
from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np

from kymata.io.cli import print_with_color, input_with_color
from kymata.io.config import load_config, modify_param_config
from kymata.io.file import PathType

_logger = getLogger(__name__)

CHANNEL_TRIGGER = 'STI101'
TRIGGER_REP_ONSET  = 3


def run_first_pass_cleansing_and_maxwell_filtering(list_of_participants: list[str],
                                                   data_root_dir: str,
                                                   dataset_directory_name: str,
                                                   n_runs: int,
                                                   emeg_machine_used_to_record_data: str,
                                                   skip_maxfilter_if_previous_runs_exist: bool,
                                                   automatic_bad_channel_detection_requested: bool,
                                                   supress_excessive_plots_and_prompts: bool,
                                                   raw_emeg_path: Optional[Path] = None,
                                                   processed_path: Optional[Path] = None,
                                                   ) -> None:

    if raw_emeg_path is None:
        raw_emeg_path = Path(data_root_dir, dataset_directory_name, "raw_emeg")
    if processed_path is None:
        processed_path = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "1_maxfiltered")
        processed_path.mkdir(exist_ok=True)

    for participant in list_of_participants:
        config_path = Path(raw_emeg_path, participant, f"{participant}_recording_config.yaml")
        for run in range(1, n_runs + 1):

            # set filename. (Use .fif.gz extension to use gzip to compress)
            saved_maxfiltered_path = Path(processed_path, f"{participant}_run{run!s}_raw_sss.fif")

            if skip_maxfilter_if_previous_runs_exist and saved_maxfiltered_path.exists():

                print_with_color(f"Skipping first pass filtering and Maxfiltering for {participant} [Run {str(run)}]...", Fore.GREEN)

            else:
                # Preprocessing Participant and run info
                print_with_color(f"Loading participant {participant} [Run {str(run)}]...", Fore.GREEN)

                # Load data
                print_with_color("   Loading Raw data...", Fore.GREEN)

                raw_fif_data = mne.io.Raw(Path(raw_emeg_path, participant, f"{participant}_run{run!s}_raw.fif"),
                                          preload=True)

                # Rename any channels that require it, and their type
                recording_config = load_config(config_path)
                ecg_and_eog_channel_name_and_type_overwrites = recording_config[
                    'ECG_and_EOG_channel_name_and_type_overwrites']

                # Set EOG and ECG types for clarity (normally shouldn't change anything, but we do it anyway)
                ecg_and_eog_channel_type_overwrites = {}
                for key, value in ecg_and_eog_channel_name_and_type_overwrites.items():
                    ecg_and_eog_channel_type_overwrites[key] = value["new_type"]
                raw_fif_data.set_channel_types(ecg_and_eog_channel_type_overwrites, verbose=None)

                # Set EOG and ECG names for clarity
                ecg_and_eog_channel_name_overwrites = {}
                for key, value in ecg_and_eog_channel_name_and_type_overwrites.items():
                    ecg_and_eog_channel_name_overwrites[key] = value["new_name"]
                raw_fif_data.rename_channels(ecg_and_eog_channel_name_overwrites, allow_duplicates=False)

                # Set bad channels (manually)
                print_with_color("   Setting bad channels...", Fore.GREEN)
                print_with_color("   ...manual", Fore.GREEN)

                raw_fif_data.info['bads'] = recording_config['bad_channels']

                response = input_with_color(
                    "Would you like to see the raw data? Recommended if you want to confirm" +
                    " ECG, HEOG, VEOG are correct, and to mark further EEG bads (they will be saved directly) " +
                    " (y/n)", Fore.MAGENTA)
                if response == "y":
                    print("...Plotting Raw data.")
                    mne.viz.plot_raw(raw_fif_data, scalings='auto', block=True)
                else:
                    print("...assuming you want to continue without looking at the raw data.")

                # Write back selected bad channels back to participant's config .yaml file
                modify_param_config(config_path,
                                    'bad_channels',
                                    [str(item) for item in sorted(raw_fif_data.info['bads'])])

                # Get the head positions
                chpi_amplitudes = mne.chpi.compute_chpi_amplitudes(raw_fif_data)
                chpi_locs = mne.chpi.compute_chpi_locs(raw_fif_data.info, chpi_amplitudes)
                head_pos_data = mne.chpi.compute_head_pos(raw_fif_data.info, chpi_locs, verbose=True)

                print_with_color("   Removing CHPI ...", Fore.GREEN)

                # Remove hpi & line
                raw_fif_data = mne.chpi.filter_chpi(raw_fif_data, include_line=False)

                print_with_color("   Removing mains component (50Hz and harmonics) from MEG & EEG...", Fore.GREEN)

                raw_fif_data.compute_psd(tmax=1000000, fmax=500, average='mean').plot()

                # note that EEG and MEG do not have the same frequencies, so we remove them seperately
                meg_picks = mne.pick_types(raw_fif_data.info, meg=True)
                meg_freqs = (50, 100, 120, 150, 200, 240, 250, 360, 400, 450)
                raw_fif_data = raw_fif_data.notch_filter(freqs=meg_freqs, picks=meg_picks)

                eeg_picks = mne.pick_types(raw_fif_data.info, eeg=True)
                eeg_freqs = (50, 150, 250, 300, 350, 400, 450)
                raw_fif_data = raw_fif_data.notch_filter(freqs=eeg_freqs, picks=eeg_picks)

                fig = raw_fif_data.compute_psd(tmax=1000000, fmax=500, average='mean').plot()
                psd_checks_dir = Path(processed_path, "power_spectral_density_checks")
                psd_checks_dir.mkdir(exist_ok=True)
                fig.savefig(Path(psd_checks_dir, f"{participant}_run{run!s}_power_spectral_density_after_notch_filters.png"))

                if automatic_bad_channel_detection_requested:
                    print_with_color("   ...automatic", Fore.GREEN)
                    raw_fif_data = apply_automatic_bad_channel_detection(raw_fif_data, emeg_machine_used_to_record_data)

                # Apply SSS and movement compensation
                print_with_color("   Applying SSS and movement compensation...", Fore.GREEN)

                fine_cal_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/sss_cal_' + emeg_machine_used_to_record_data + '.dat'))
                crosstalk_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/ct_sparse_' + emeg_machine_used_to_record_data + '.fif'))
                
                if not supress_excessive_plots_and_prompts:
                    mne.viz.plot_head_positions(
                        head_pos_data, mode='field', destination=raw_fif_data.info['dev_head_t'], info=raw_fif_data.info)

                raw_fif_data_sss_movecomp_tr = mne.preprocessing.maxwell_filter(
                    raw_fif_data,
                    cross_talk=crosstalk_file,
                    calibration=fine_cal_file,
                    head_pos=head_pos_data,
                    coord_frame='head',
                    st_correlation=0.980,
                    st_duration=10,
                    destination=(0, 0, 0.04),
                    # note that using extended_proj/eSSS makes no difference
                    verbose=True)

                fig = raw_fif_data_sss_movecomp_tr.compute_psd(tmax=1000000, fmax=500, average='mean').plot()
                fig.savefig(Path(psd_checks_dir, f"{participant}_run{run!s}_power_spectral_density_aftermaxfilter.png"))

                raw_fif_data_sss_movecomp_tr.save(saved_maxfiltered_path, fmt='short')


def run_second_pass_cleansing_and_eog_removal(list_of_participants: list[str],
                                              data_root_dir: str,
                                              dataset_directory_name: str,
                                              n_runs: int,
                                              remove_ecg: bool,
                                              remove_veoh_and_heog: bool,
                                              skip_ica_if_previous_runs_exist: bool,
                                              supress_excessive_plots_and_prompts: bool,
                                              ):

    cleaned_dir = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "2_cleaned")
    cleaned_dir.mkdir(exist_ok=True)

    for participant in list_of_participants:
        for run in range(1, n_runs + 1):

            saved_cleaned_path = Path(cleaned_dir, participant, "_run", str(run), '_cleaned_raw.fif.gz')
            
            if skip_ica_if_previous_runs_exist and os.path.isfile(saved_cleaned_path):

                print_with_color(f"Skipping second pass cleaning (ICA) for {participant} [Run {str(run)}]...", Fore.GREEN)

            else:

                # Preprocessing Participant and run info
                print_with_color(f"Loading participant {participant} [Run {str(run)}]...", Fore.GREEN)

                # set filename. (Use .fif.gz extension to use gzip to compress)
                saved_maxfiltered_filename = data_root_dir + dataset_directory_name + '/interim_preprocessing_files/1_maxfiltered/' + participant + "_run" + str(
                    run) + '_raw_sss.fif'

                # Load data
                print_with_color("   Loading Raw data...", Fore.GREEN)

                raw_fif_data_sss_movecomp_tr = mne.io.Raw(saved_maxfiltered_filename, preload=True)

                if not supress_excessive_plots_and_prompts:
                    response = input_with_color(
                        "Would you like to see the SSS, movement compensated, raw data data? (y/n)",
                        Fore.MAGENTA)
                    if response == "y":
                        print("...Plotting Raw data.")
                        mne.viz.plot_raw(raw_fif_data_sss_movecomp_tr, block=True)
                    else:
                        print("[y] not pressed. Assuming you want to continue without looking at the raw data.")

                # EEG channel interpolation
                print_with_color("   Interpolating EEG...", Fore.GREEN)

                print("Bads channels: " + str(raw_fif_data_sss_movecomp_tr.info["bads"]))

                # Channels marked as “bad” have been effectively repaired by SSS,
                # eliminating the need to perform MEG interpolation.

                raw_fif_data_sss_movecomp_tr = raw_fif_data_sss_movecomp_tr.interpolate_bads(reset_bads=True,
                                                                                            mode='accurate')

                # Use common average reference, not the nose reference.
                print_with_color("   Use common average EEG reference...", Fore.GREEN)

                raw_fif_data_sss_movecomp_tr = raw_fif_data_sss_movecomp_tr.set_eeg_reference(ref_channels='average')

                # remove very slow drift
                print_with_color("   Removing slow drift...", Fore.GREEN)
                raw_fif_data_sss_movecomp_tr = raw_fif_data_sss_movecomp_tr.filter(l_freq=0.1, h_freq=None, picks=None)

                # Remove ECG, VEOH and HEOG

                if remove_ecg or remove_veoh_and_heog:

                    # remove both frequencies faster than 40Hz and slow drift less than 1hz
                    filt_raw = raw_fif_data_sss_movecomp_tr.copy().filter(l_freq=1., h_freq=40)

                    ica = mne.preprocessing.ICA(n_components=30, method='fastica', max_iter='auto', random_state=97)
                    ica.fit(filt_raw)

                    explained_var_ratio = ica.get_explained_variance_ratio(filt_raw)
                    for channel_type, ratio in explained_var_ratio.items():
                        print(
                            f'   Fraction of {channel_type} variance explained by all components: '
                            f'   {ratio}'
                        )

                    ica.exclude = []

                    if supress_excessive_plots_and_prompts:
                         _remove_ecg_eog(filt_raw, ica)

                    else:
                        if remove_ecg:
                            _remove_ecg(filt_raw, ica)

                        if remove_veoh_and_heog:
                            _remove_veoh_and_heog(filt_raw, ica)

                    ica.apply(raw_fif_data_sss_movecomp_tr)

                    if not supress_excessive_plots_and_prompts:
                        mne.viz.plot_raw(raw_fif_data_sss_movecomp_tr)

                raw_fif_data_sss_movecomp_tr.save(Path(cleaned_dir, participant + "_run" + str(run) + '_cleaned_raw.fif.gz'),
                    overwrite=True)


def _remove_veoh_and_heog(filt_raw, ica):
    """
    Note: mutates `ica`.
    """
    print("   ...Starting EOG removal.")
    eog_evoked = mne.preprocessing.create_eog_epochs(filt_raw).average()
    eog_evoked.apply_baseline(baseline=(None, -0.2))
    eog_evoked.plot_joint()
    # find which ICs match the EOG pattern
    eog_indices, eog_scores = ica.find_bads_eog(filt_raw)
    ica.exclude += eog_indices
    # plot ICs applied to raw data, with EOG matches highlighted
    ica.plot_sources(filt_raw, show_scrollbars=True)
    # barplot of ICA component "EOG match" scores
    ica.plot_scores(eog_scores)
    # plot diagnostics
    ica.plot_properties(filt_raw, picks=eog_indices)
    # plot ICs applied to the averaged EOG epochs, with EOG matches highlighted
    ica.plot_sources(eog_evoked)
    # blinks
    ica.plot_overlay(filt_raw, exclude=eog_indices, picks='eeg')


def _remove_ecg(filt_raw, ica):
    """
    Note: mutates `ica`.
    """
    print_with_color("   Starting ECG removal...", Fore.GREEN)
    ecg_evoked = mne.preprocessing.create_ecg_epochs(filt_raw).average()
    ecg_evoked.apply_baseline(baseline=(None, -0.2))
    ecg_evoked.plot_joint()
    # find which ICs match the ECG pattern
    ecg_indices, ecg_scores = ica.find_bads_ecg(filt_raw)
    ica.exclude = ecg_indices
    # plot ICs applied to raw data, with ECG matches highlighted
    ica.plot_sources(filt_raw, show_scrollbars=True)
    # barplot of ICA component "ECG match" scores
    ica.plot_scores(ecg_scores)
    # plot diagnostics
    ica.plot_properties(filt_raw, picks=ecg_indices)
    # plot ICs applied to the averaged ECG epochs, with ECG matches highlighted
    ica.plot_sources(ecg_evoked)
    # heartbeats
    ica.plot_overlay(filt_raw, exclude=ica.exclude, picks='mag')


def _remove_ecg_eog(filt_raw, ica):
    """
    Note: mutates `ica`.
    """
    print_with_color("   Starting ECG and EOG removal...", Fore.GREEN)
    ecg_indices, ecg_scores = ica.find_bads_ecg(filt_raw)
    eog_indices, eog_scores = ica.find_bads_eog(filt_raw)
    ica.exclude = ecg_indices
    ica.exclude += eog_indices
    eog_scores.append(ecg_scores)
    ica.plot_sources(filt_raw, show_scrollbars=True)
    ica.plot_scores(eog_scores)


def estimate_noise_cov(data_root_dir: str,
                       emeg_machine_used_to_record_data: str,
                       list_of_participants: list[str],
                       dataset_directory_name: str,
                       n_runs: int,
                       cov_method: str,
                       duration_emp,
                       reg_method,
                       ):

    emeg_dir = Path(data_root_dir, dataset_directory_name, "raw_emeg")

    cleaned_dir = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "2_cleaned")
    cleaned_dir.mkdir(exist_ok=True)

    sensor_data_dir = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "4_hexel_current_reconstruction")
    sensor_data_dir.mkdir(exist_ok=True)

    cov_dir = Path(sensor_data_dir, "noise_covariance_files")
    cov_dir.mkdir(exist_ok=True)

    for p in list_of_participants:
        if cov_method == 'grandave':
            cleaned_raws = []
            for run in range(1, n_runs + 1):
                raw_fname = Path(cleaned_dir, p + '_run' + str(run) + '_cleaned_raw.fif.gz')
                raw = mne.io.Raw(raw_fname, preload=True)
                raw_cropped = raw.crop(tmin=0, tmax=800)
                cleaned_raws.append(raw_cropped)
            raw_combined = mne.concatenate_raws(raws=cleaned_raws, preload=True)
            raw_epoch = mne.make_fixed_length_epochs(raw_combined, duration=800, preload=True, reject_by_annotation=False)
            cov = mne.compute_covariance(raw_epoch, tmin=0, tmax=None, method=reg_method, return_estimators=True)
            mne.write_cov(data_root_dir + dataset_directory_name + '/interim_preprocessing_files/3_evoked_sensor_data/covariance_grand_average/' + p + '-grandave-cov.fif', cov)
            
        elif cov_method == 'emptyroom':
            raw_fname = Path(cleaned_dir, p + '_run1' + '_cleaned_raw.fif.gz')
            raw = mne.io.Raw(raw_fname, preload=True)
            emptyroom_fname = data_root_dir + dataset_directory_name + '/raw_emeg/' + p + '/' + p + '_empty_room.fif'
            emptyroom_raw = mne.io.Raw(emptyroom_fname, preload=True)
            emptyroom_raw = mne.preprocessing.maxwell_filter_prepare_emptyroom(emptyroom_raw, raw=raw)

            fine_cal_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/sss_cal_' + emeg_machine_used_to_record_data + '.dat'))
            crosstalk_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/ct_sparse_' + emeg_machine_used_to_record_data + '.fif'))
  
            raw_fif_data_sss = mne.preprocessing.maxwell_filter(
                        emptyroom_raw,
                        cross_talk=crosstalk_file,
                        calibration=fine_cal_file,
                        st_correlation=0.980,
                        st_duration=10,
                        # note that using extended_proj/eSSS makes no difference
                        verbose=True)

            cov = mne.compute_raw_covariance(raw_fif_data_sss, tmin=0, tmax=duration_emp, method=reg_method, return_estimators=True)
            if duration_emp is None:
                mne.write_cov(Path(cov_dir, p + '-emptyroom-cov.fif'), cov)
            else:
                mne.write_cov(Path(cov_dir, p + '-emptyroom' + str(duration_emp) + '-cov.fif'), cov)
        
        elif cov_method == 'runstart':
            cleaned_raws = []
            for run in range(1, n_runs + 1):
                raw_fname = Path(cleaned_dir, p + '_run' + str(run) + '_cleaned_raw.fif.gz')
                raw = mne.io.Raw(raw_fname, preload=True)
                raw_cropped = raw.crop(tmin=0, tmax=20)
                cleaned_raws.append(raw_cropped)
            raw_combined = mne.concatenate_raws(raws=cleaned_raws, preload=True)
            raw_epoch = mne.make_fixed_length_epochs(raw_combined, duration=20, preload=True, reject_by_annotation=False)
            cov = mne.compute_covariance(raw_epoch, tmin=0, tmax=None, method=reg_method, return_estimators=True)
            mne.write_cov(data_root_dir + dataset_directory_name + '/interim_preprocessing_files/3_evoked_sensor_data/covariance_grand_average/' + p + '-runstart-cov.fif', cov)

        elif cov_method == 'fusion':

            # First calculate the covariance for EEG using grandave
            cleaned_raws = []
            for run in range(1, n_runs + 1):
                raw_fname = Path(cleaned_dir, p + '_run' + str(run) + '_cleaned_raw.fif.gz')
                raw = mne.io.Raw(raw_fname, preload=True)
                raw_cropped = raw.crop(tmin=0, tmax=800)
                cleaned_raws.append(raw_cropped)
            raw_combined = mne.concatenate_raws(raws=cleaned_raws, preload=True)
            raw_epoch = mne.make_fixed_length_epochs(raw_combined, duration=800, preload=True, reject_by_annotation=False)
            cov_eeg = mne.compute_covariance(raw_epoch, tmin=0, tmax=None, method=reg_method, return_estimators=True)
            del cleaned_raws, raw_combined, raw_epoch

            # Now calcualte the covariance for MEG using emptyroom
            emptyroom_fname = Path(emeg_dir, p, p + '_empty_room_raw.fif')
            emptyroom_raw = mne.io.Raw(emptyroom_fname, preload=True)
            emptyroom_raw = mne.preprocessing.maxwell_filter_prepare_emptyroom(emptyroom_raw, raw=raw)

            fine_cal_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/sss_cal_' + emeg_machine_used_to_record_data + '.dat'))
            crosstalk_file = str(Path(Path(__file__).parent.parent.parent, 'kymata-core-data', 'cbu_specific_files/SSS/ct_sparse_' + emeg_machine_used_to_record_data + '.fif'))
  
            raw_fif_data_sss = mne.preprocessing.maxwell_filter(
                        emptyroom_raw,
                        cross_talk=crosstalk_file,
                        calibration=fine_cal_file,
                        st_correlation=0.980,
                        st_duration=10,
                        # note that using extended_proj/eSSS makes no difference
                        verbose=True)

            cov_meg = mne.compute_raw_covariance(raw_fif_data_sss, tmin=0, tmax=1, method=reg_method, return_estimators=True)
            del raw, emptyroom_raw, raw_fif_data_sss

            # Now combine the two covariance matrices
            cov_data = cov_eeg.data
            meg_indices_ave = [index for index, channel in enumerate(cov_eeg.ch_names) if 'EEG' not in channel]
            meg_indices_emp = [index for index, channel in enumerate(cov_meg.ch_names) if 'EEG' not in channel]
            cov_data[np.ix_(meg_indices_ave, meg_indices_ave)] = cov_meg.data[np.ix_(meg_indices_emp, meg_indices_emp)]
            cov = mne.Covariance(cov_data, names=cov_eeg.ch_names, bads=cov_eeg['bads'], projs=cov_eeg['projs'], nfree=cov_eeg.nfree)

            mne.write_cov(Path(cov_dir, p + '-fusion-cov.fif'), cov)


def create_trialwise_data(data_root_dir: PathType,
                          dataset_directory_name: str,
                          list_of_participants: list[str],
                          repetitions_per_runs: int,
                          stimulus_length: int,  # seconds
                          number_of_runs: int,
                          latency_range: tuple[float, float]  # seconds
                          ):
    """Create trials objects from the raw data files (still in sensor space)"""

    cleaned_dir = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "2_cleaned")

    trialwise_sensorspace_dir = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files", "3_trialwise_sensorspace")
    trialwise_sensorspace_dir.mkdir(exist_ok=True)

    evoked_path = Path(trialwise_sensorspace_dir, "evoked_data")
    evoked_path.mkdir(exist_ok=True)

    logs_path = Path(trialwise_sensorspace_dir, "logs")
    logs_path.mkdir(exist_ok=True)

    print(f"{Fore.GREEN}{Style.BRIGHT}Starting trials and {Style.RESET_ALL}")

    for p in list_of_participants:

        print(f"{Fore.GREEN}{Style.BRIGHT}...Concatenating trials{Style.RESET_ALL}")

        cleaned_raws = []

        for run in range(1, number_of_runs + 1):
            raw_path = Path(cleaned_dir, f'{p}_run{run}_cleaned_raw.fif.gz')
            print(f"Loading {raw_path}...")
            if os.path.isfile(raw_path):
                raw = mne.io.Raw(raw_path, preload=True)
                cleaned_raws.append(raw)

        raw: mne.io.Raw = mne.io.concatenate_raws(raws=cleaned_raws, preload=True)

        raw_events = mne.find_events(raw, stim_channel=CHANNEL_TRIGGER, shortest_event=1)
        repetition_events = mne.pick_events(raw_events, include=TRIGGER_REP_ONSET)
        # name repetitions
        for i in range(len(repetition_events)):
            repetition_events[i][2] = str(i)

        # Validate repetition events
        assert len(repetition_events) == number_of_runs * repetitions_per_runs, \
            f"{len(repetition_events)=} but {number_of_runs * repetitions_per_runs=}"

        # Denote picks
        include = []  # ['MISC006']  # MISC05, trigger channels etc, if needed
        picks: NDArray = mne.pick_types(raw.info, meg=True, eeg=True, stim=False, exclude='bads', include=include)
        _logger.info(f"Picked {picks.shape} channels out of {len(raw.info['ch_names'])}")

        print(f"{Fore.GREEN}{Style.BRIGHT}... extract and save evoked data{Style.RESET_ALL}")

        # Extract trial instances ('epochs')
        _tmin = latency_range[0]
        _tmax = (stimulus_length
                 # extra padding at the end to accommodate range of latencies
                 + latency_range[1]
                 # Extra space to account for audio latency drift
                 + 2)

        epochs = mne.Epochs(raw, repetition_events, None, _tmin, _tmax, picks=picks, baseline=(None, None), preload=True)
        _logger.info(f"Created epochs with {len(epochs.ch_names)} channels")

        # Log which channels are worst
        dropfig = epochs.plot_drop_log(subject=p)
        dropfig.savefig(Path(logs_path, f"drop-log_{p}.jpg"))

        # Save individual repetitions
        for i in range(len(repetition_events)):
            evoked = epochs[str(i)].average()
            _logger.info(f"Individual evokeds created with {len(evoked.ch_names)} channels (i.e. {evoked.data.shape=})")
            evoked.save(Path(evoked_path, f"{p}_rep{i}.fif"), overwrite=True)

        # Average over repetitions
        evoked = epochs.average()
        _logger.info(f"Average evokeds created with {len(evoked.ch_names)} channels (i.e. {evoked.data.shape=})")

        evoked.save(Path(evoked_path, f"{p}-ave.fif"), overwrite=True)


def apply_automatic_bad_channel_detection(raw_fif_data: mne.io.Raw, machine_used: str, plot: bool = True):
    """Apply Automatic Bad Channel Detection"""
    raw_check = raw_fif_data.copy()

    fine_cal_file = Path(Path(__file__).parent.parent,
                         'kymata-core-data', 'cbu_specific_files',
                         'sss_cal' + machine_used + '.dat')
    crosstalk_file = Path(Path(__file__).parent.parent,
                          'kymata-core-data', 'cbu_specific_files',
                          'ct_sparse' + machine_used + '.fif')

    auto_noisy_chs, auto_flat_chs, auto_scores = mne.preprocessing.find_bad_channels_maxwell(
        raw_check, cross_talk=crosstalk_file, calibration=fine_cal_file,
        return_scores=True, verbose=True)
    print(auto_noisy_chs)
    print(auto_flat_chs)

    bads = raw_fif_data.info['bads'] + auto_noisy_chs + auto_flat_chs
    raw_fif_data.info['bads'] = bads

    if plot:
        _plot_bad_chans(auto_scores)

    return raw_fif_data


def _plot_bad_chans(auto_scores):

    # Only select the data for gradiometer channels.
    ch_type = 'grad'
    ch_subset = auto_scores['ch_types'] == ch_type
    ch_names = auto_scores['ch_names'][ch_subset]
    scores = auto_scores['scores_noisy'][ch_subset]
    limits = auto_scores['limits_noisy'][ch_subset]
    bins = auto_scores['bins']  # The windows that were evaluated.

    # We will label each segment by its start and stop time, with up to 3
    # digits before and 3 digits after the decimal place (1 ms precision).
    bin_labels = [f'{start:3.3f} – {stop:3.3f}'
                  for start, stop in bins]

    # We store the data in a Pandas DataFrame. The seaborn heatmap function
    # we will call below will then be able to automatically assign the correct
    # labels to all axes.
    data_to_plot = DataFrame(data=scores,
                             columns=Index(bin_labels, name='Time (s)'),
                             index=Index(ch_names, name='Channel'))

    # First, plot the "raw" scores.
    fig, ax = plt.subplots(1, 2, figsize=(12, 8))
    fig.suptitle(f'Automated noisy channel detection: {ch_type}',
                 fontsize=16, fontweight='bold')
    sns.heatmap(data=data_to_plot, cmap='Reds', cbar_kws=dict(label='Score'),
                ax=ax[0])
    [ax[0].axvline(x, ls='dashed', lw=0.25, dashes=(25, 15), color='gray')
     for x in range(1, len(bins))]
    ax[0].set_title('All Scores', fontweight='bold')

    # Now, adjust the color range to highlight segments that exceeded the limit.
    sns.heatmap(data=data_to_plot,
                vmin=nanmin(limits),  # bads in input data have NaN limits
                cmap='Reds', cbar_kws=dict(label='Score'), ax=ax[1])
    [ax[1].axvline(x, ls='dashed', lw=0.25, dashes=(25, 15), color='gray')
     for x in range(1, len(bins))]
    ax[1].set_title('Scores > Limit', fontweight='bold')

    # The figure title should not overlap with the subplots.
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    # Replace the word “noisy” with “flat”, and replace
    # vmin=nanmin(limits) with vmax=nanmax(limits) to print flat channels



import numpy as np
from numpy.typing import NDArray


def normalize(x: NDArray, inplace: bool = False) -> NDArray:
    """
    Remove the mean and divide by the Euclidean magnitude.
    If inplace is True, the array will be modified in place. If false, a normalized copy will be returned.
    raises: ZeroDivisionError
    """
    if not inplace:
        x = np.copy(x)

    x -= np.mean(x, axis=-1, keepdims=True)

    # In case the values of x are very small, sometimes _magnitude can return 0, which would cause a divide by zero
    # error. Having already centred x, we can upscale it before downscaling it to avoid this issue.
    # If the _magnitude should actually be 0 (i.e. an error), this won't make a difference to that.
    if (_normalize_magnitude(x) == 0).any():
        x *= 1_000_000
    # If we STILL have a magnitude-0 vector, we will have a problem, so should raise the error immediately.
    with np.errstate(divide="raise"):
        x /= _normalize_magnitude(x)

    return x


def _normalize_magnitude(x: NDArray) -> NDArray:
    """Reusable magnitude function for use in `normalize`."""
    return np.sqrt(np.sum(x**2, axis=-1, keepdims=True))


def get_stds(x: NDArray, n: int):
    """
    Get the stds (times sqrt(n)) to correct for normalisation difference between each set of n_samples
    """
    d0, d1, d2 = x.shape
    y = np.concatenate((np.zeros((d0, d1, 1)), np.cumsum(x**2, axis=-1)), axis=-1)
    z = np.concatenate((np.zeros((d0, d1, 1)), np.cumsum(x, axis=-1)), axis=-1)
    y = y[:,:,-n-1:-1] - y[:,:,:n]
    z = z[:,:,-n-1:-1] - z[:,:,:n]
    return (y - ((z**2) / n)) ** 0.5



"""
# Version 0.4: allowed mismatched hexel ids in the left and right hemispheres
#
# Version 0.3: moved to log p-values.
#
# Version 0.2: supports sensor data.
#
# Version 0.1: original layered sparse data format.
"""

from io import TextIOWrapper
from pathlib import Path
from typing import Any
from zipfile import ZipFile

from numpy import frombuffer
from numpy.typing import NDArray

from kymata.entities.datatypes import LatencyDType, FunctionNameDType, HexelDType
from kymata.io.file import PathType, FileType, open_or_use


# noinspection DuplicatedCode
def _load_data_0_3(from_path_or_file: PathType | FileType) -> dict[str, Any]:
    """
    This is a function which loads data format 0.3.

    The idea is that *this function never changes*, and handling code instead can change to
    transform the returned dictionary to whatever format is required for the current
    iteration of the appropriate classes.

    In particular, that means it uses inline string keys rather than keys stored in the Keys class.
    """
    return_dict = dict()
    with open_or_use(from_path_or_file, mode="rb") as archive, ZipFile(archive, "r") as zf:

        with TextIOWrapper(zf.open("_metadata/expression-set-type.txt"), encoding="utf-8") as f:
            return_dict["expressionset-type"] = str(f.read()).strip()
        with TextIOWrapper(zf.open("/layers.txt"), encoding="utf-8") as f:
            layers = [str(line.strip()) for line in f.readlines()]
        with TextIOWrapper(zf.open("/channels.txt"), encoding="utf-8") as f:
            return_dict["channels"] = [c.strip() for c in f.readlines()]
        with TextIOWrapper(zf.open("/latencies.txt"), encoding="utf-8") as f:
            return_dict["latencies"] = [LatencyDType(lat.strip()) for lat in f.readlines()]
        with TextIOWrapper(zf.open("/functions.txt"), encoding="utf-8") as f:
            return_dict["functions"] = [FunctionNameDType(fun.strip()) for fun in f.readlines()]
        return_dict["data"] = dict()
        for layer in layers:
            with zf.open(f"/{layer}/coo-coords.bytes") as f:
                coords: NDArray = frombuffer(f.read(), dtype=int).reshape((3, -1))
            with zf.open(f"/{layer}/coo-data.bytes") as f:
                data: NDArray = frombuffer(f.read(), dtype=float)
            with TextIOWrapper(zf.open(f"/{layer}/coo-shape.txt"), encoding="utf-8") as f:
                shape: tuple[int, ...] = tuple(int(s.strip()) for s in f.readlines())
            return_dict["data"][layer] = dict()
            return_dict["data"][layer]["coords"] = coords
            return_dict["data"][layer]["shape"] = shape
            return_dict["data"][layer]["data"] = data
    return return_dict


# noinspection DuplicatedCode
def _load_data_0_2(from_path_or_file: PathType | FileType) -> dict[str, Any]:
    """
    This is a function which loads data format 0.2.

    The idea is that *this function never changes*, and handling code instead can change to
    transform the returned dictionary to whatever format is required for the current
    iteration of the appropriate classes.

    In particular, that means it uses inline string keys rather than keys stored in the Keys class.
    """
    return_dict = dict()
    with open_or_use(from_path_or_file, mode="rb") as archive, ZipFile(archive, "r") as zf:

        with TextIOWrapper(zf.open("_metadata/expression-set-type.txt"), encoding="utf-8") as f:
            return_dict["expressionset-type"] = str(f.read()).strip()
        with TextIOWrapper(zf.open("/layers.txt"), encoding="utf-8") as f:
            layers = [str(line.strip()) for line in f.readlines()]
        with TextIOWrapper(zf.open("/channels.txt"), encoding="utf-8") as f:
            return_dict["channels"] = [c.strip() for c in f.readlines()]
        with TextIOWrapper(zf.open("/latencies.txt"), encoding="utf-8") as f:
            return_dict["latencies"] = [LatencyDType(lat.strip()) for lat in f.readlines()]
        with TextIOWrapper(zf.open("/functions.txt"), encoding="utf-8") as f:
            return_dict["functions"] = [FunctionNameDType(fun.strip()) for fun in f.readlines()]
        return_dict["data"] = dict()
        for layer in layers:
            with zf.open(f"/{layer}/coo-coords.bytes") as f:
                coords: NDArray = frombuffer(f.read(), dtype=int).reshape((3, -1))
            with zf.open(f"/{layer}/coo-data.bytes") as f:
                data: NDArray = frombuffer(f.read(), dtype=float)
            with TextIOWrapper(zf.open(f"/{layer}/coo-shape.txt"), encoding="utf-8") as f:
                shape: tuple[int, ...] = tuple(int(s.strip()) for s in f.readlines())
            return_dict["data"][layer] = dict()
            return_dict["data"][layer]["coords"] = coords
            return_dict["data"][layer]["shape"] = shape
            return_dict["data"][layer]["data"] = data
    return return_dict


# noinspection DuplicatedCode
def _load_data_0_1(from_path_or_file: PathType | FileType) -> dict[str, Any]:
    """
    This is a function which loads data format 0.1.

    The idea is that *this function never changes*, and handling code instead can change to
    transform the returned dictionary to whatever format is required for the current
    iteration of the appropriate classes.

    In particular, that means it uses inline string keys rather than keys stored in the Keys class.
    """

    if isinstance(from_path_or_file, str):
        from_path_or_file = Path(from_path_or_file)

    return_dict = dict()
    with open_or_use(from_path_or_file, mode="rb") as archive, ZipFile(archive, "r") as zf:
        with TextIOWrapper(zf.open("/hexels.txt"), encoding="utf-8") as f:
            return_dict["hexels"]: list[HexelDType] = [HexelDType(h.strip()) for h in f.readlines()]
        with TextIOWrapper(zf.open("/latencies.txt"), encoding="utf-8") as f:
            return_dict["latencies"]: list[LatencyDType] = [LatencyDType(lat.strip()) for lat in f.readlines()]
        with TextIOWrapper(zf.open("/functions.txt"), encoding="utf-8") as f:
            return_dict["functions"]: list[FunctionNameDType] = [FunctionNameDType(fun.strip()) for fun in f.readlines()]
        return_dict["data"] = dict()
        for layer in ["left", "right"]:
            with zf.open(f"/{layer}/coo-coords.bytes") as f:
                coords: NDArray = frombuffer(f.read(), dtype=int).reshape((3, -1))
            with zf.open(f"/{layer}/coo-data.bytes") as f:
                data: NDArray = frombuffer(f.read(), dtype=float)
            with TextIOWrapper(zf.open(f"/{layer}/coo-shape.txt"), encoding="utf-8") as f:
                shape: tuple[int, ...] = tuple(int(s.strip()) for s in f.readlines())
            return_dict["data"][layer] = dict()
            return_dict["data"][layer]["coords"] = coords
            return_dict["data"][layer]["shape"] = shape
            return_dict["data"][layer]["data"] = data

    return return_dict



from pathlib import Path

import yaml

from kymata.datasets.data_root import data_root_path
from kymata.io.file import PathType, FileType, open_or_use


def load_config(config_location: PathType | FileType):
    """
    Load configuration parameters from a specified path or file.

    This function reads the configuration parameters from a YAML file located at the given path or open file.

    Args:
        config_location (PathType | FileType): The path to the configuration file or an open file object.

    Returns:
        dict: The configuration parameters loaded from the file.

    Raises:
        FileNotFoundError: If the specified path does not exist.
        yaml.YAMLError: If there is an error in parsing the YAML file.
    """
    with open_or_use(config_location) as stream:
        return yaml.safe_load(stream)


def modify_param_config(config_location: str, key: str, value):
    """
    Modify a specific configuration parameter in the given configuration file.

    This function updates the value of a specified key in the configuration file and saves the changes.

    Args:
        config_location (str): The path to the configuration file.
        key (str): The key of the configuration parameter to be modified.
        value: The new value to be assigned to the specified key.

    Raises:
        FileNotFoundError: If the specified configuration file does not exist.
        yaml.YAMLError: If there is an error in parsing the YAML file.
    """
    with open(config_location, 'r') as stream:
        """Load config parameter"""
        data = yaml.safe_load(stream)
        data[key] = value
    yaml.emitter.Emitter.process_tag = lambda self, *args, **kw: None
    with open(config_location, 'w') as file:
        yaml.dump(data,
                  file,
                  sort_keys = False)


def get_root_dir(config: dict) -> str:
    """
    Get the root directory based on the configuration parameters.

    This function returns the appropriate root directory path based on the 'data_location' parameter
    in the provided configuration dictionary.

    Args:
        config (dict): The configuration dictionary containing the 'data_location' parameter.

    Returns:
        str: The root directory path corresponding to the 'data_location' parameter.

    Raises:
        ValueError: If the 'data_location' parameter is not 'local', 'cbu', or 'cbu-local'.
    """
    if config['data_location'] == "local":
        return str(Path(data_root_path(), "emeg_study_data")) + "/"
    elif config['data_location'] == "cbu":
        return '/imaging/projects/cbu/kymata/data/'
    elif config['data_location'] == "cbu-local":
        return '//cbsu/data/imaging/projects/cbu/kymata/data/'
    else:
        raise ValueError(
            "The `data_location` parameter in the config file must be either 'cbu' or 'local' or 'cbu-local'.")



from copy import deepcopy
from typing import Dict, List, Tuple

import numpy as np

from kymata.ippm.data_tools import IPPMHexel, Node


class IPPMBuilder:
    def __init__(
            self,
            hexels: Dict[str, IPPMHexel],
            inputs: List[str],
            hierarchy: Dict[str, List[str]],
            hemisphere: str
    ):
        self._hexels = deepcopy(hexels)
        self._inputs = inputs
        self._hierarchy = deepcopy(hierarchy)
        self._hemisphere = hemisphere
        self._graph = {}  # Chosen structure is Dict[str, Node] since that enables quick look up for node position

    def build_graph_dict(self) -> Dict[str, Node]:
        self._hexels = self._sort_hexel_spikes_by_latency_asc()

        y_axis_partition_size = 1 / len(self._hierarchy.keys()) if len(self._hierarchy.keys()) > 0 else 1
        partition_ptr = 0
        while childless_functions := self._get_childless_functions():
            for childless_func in childless_functions:
                self._graph = self._create_nodes_and_edges_for_function(childless_func, partition_ptr, y_axis_partition_size)
                self._hierarchy.pop(childless_func)
                partition_ptr += 1

        return self._graph

    def _sort_hexel_spikes_by_latency_asc(self) -> Dict[str, IPPMHexel]:
        for function in self._hexels.keys():
            if self._hemisphere == "rightHemisphere":
                self._hexels[function].right_best_pairings.sort(key=lambda x: x[0])
            else:
                self._hexels[function].left_best_pairings.sort(key=lambda x: x[0])
        return self._hexels

    def _get_childless_functions(self):
        def __unpack_dict_values_into_list(dict_to_unpack):
            return [value for values in dict_to_unpack.values() for value in values]

        current_functions = set(self._hierarchy.keys())
        functions_with_children = set(__unpack_dict_values_into_list(self._hierarchy))
        # When no functions left, it returns empty set.
        return current_functions.difference(functions_with_children)

    def _create_nodes_and_edges_for_function(
            self,
            function_name: str,
            partition_ptr: int,
            partition_size: float
    ) -> Dict[str, Node]:
        def __get_y_coordinate(curr_partition_number: int, partition_size: float) -> float:
            return 1 - partition_size * curr_partition_number

        func_parents = self._hierarchy[function_name]
        current_y_axis_coord = __get_y_coordinate(partition_ptr, partition_size)
        if function_name in self._inputs:
            self._graph[function_name] = Node(100, (0, current_y_axis_coord), [])
        else:
            childless_func_pairings = self._get_best_pairings_from_hemisphere(function_name)
            self._graph = self._create_nodes_for_childless_function(
                current_y_axis_coord,
                childless_func_pairings,
                function_name
            )
            self._graph = self._create_edges_between_parents_and_childless_function(
                func_parents,
                function_name
            )

        return self._graph

    def _get_best_pairings_from_hemisphere(self, func: str) -> List[Tuple[float, float]]:
        if func in self._hexels.keys():
            return (self._hexels[func].right_best_pairings if self._hemisphere == "rightHemisphere"
                    else self._hexels[func].left_best_pairings)
        return []

    def _create_nodes_for_childless_function(
            self,
            current_y_axis_coord: float,
            childless_func_pairings: List[Tuple[float, float]],
            function_name: str,
    ):
        def __map_magnitude_to_node_size(magnitude: float) -> float:
            return -10 * np.log10(magnitude)

        for idx, pair in enumerate(childless_func_pairings):
            latency, magnitude = pair
            parent_function = [function_name + '-' + str(idx - 1)] if idx != 0 else []
            self._graph[function_name + '-' + str(idx)] = Node(__map_magnitude_to_node_size(magnitude),
                                                         (latency, current_y_axis_coord),
                                                         parent_function)

        return self._graph

    def _create_edges_between_parents_and_childless_function(
            self,
            parents: List[str],
            function_name: str,
    ) -> Dict[str, Node]:
        if function_name in self._hexels.keys():
            for parent in parents:
                if parent in self._inputs:
                    self._graph[function_name + "-0"].inc_edges.append(parent)
                else:
                    parent_pairings = self._get_best_pairings_from_hemisphere(parent)
                    if len(parent_pairings) > 0:
                        self._graph[function_name + "-0"].inc_edges.append(parent + '-' + str(len(parent_pairings) - 1))

        return self._graph



from copy import deepcopy
from typing import List, Dict
from collections import namedtuple

import numpy as np

from .data_tools import IPPMHexel

# convenient tuple/class to hold information about nodes.
Node = namedtuple('Node', 'magnitude position in_edges')


class IPPMBuilder(object):
    """
    A graphing class used to construct a dictionary that contains the nodes and all relevant
    information to construct a dict containing node names as keys and Node objects (see namedtuple) as values.
    """

    def build_graph(self,
                    hexels: Dict[str, IPPMHexel],
                    function_hier : Dict[str, List[str]],
                    inputs : List[str],
                    hemi : str) -> Dict[str, Node]:
        """
        Builds a dictionary of nodes and information about the node. The information
        is built out of namedtuple class Node, which contains magnitude, position, color, and
        the incoming edges.

        Args:
            hexels (Dict[str, IPPMHexel]): Dictionary containing function names and Hexel objects with data in it.
            function_hier (Dict[str, List[str]]): Dictionary of the format {function_name: [parent_functions]}.
            inputs (List[str]): List of input functions. function_hier contains the input functions, so we need this
                                to distinguish between inputs and functions.
            hemi (str): 'leftHemisphere' or 'rightHemisphere'.

        Returns:
            Dict[str, Node]: A dictionary of nodes with unique names where the keys are node objects with all
                             relevant information for plotting a graph.

        Analysis:
            The overall complexity of the algorithm is O(f * nlogn) where:

             - f is the number of functions.
             - n is the number of hexels.

            The sorting step takes O(f * nlogn).

            Looping through functions and parents contributes O(f * n + f^2).

            Space Complexity:

             - Total space complexity is O(f * n), which includes copies of hexels and function hierarchy.

            The analysis section provides detailed reasoning for the complexities.

        Algorithm:
            The algorithm iteratively selects the top-level function, defined as a function that does not
            have any children (no outgoing arrows). It starts with the final function and proceeds
            in a top-down fashion towards the input node. Each selected top-level function creates a spike for every
            pairing in the best_pairings. Edges are added from the final function of each parent to the first current
            function, repeating for all functions. The input node is treated differently with a default size of 10 at
            latency 0.
        """

        hexels = deepcopy(hexels) # do it in-place to avoid modifying hexels.
        
        # sort it so that the null edges go in the right order (from left to right along time axis)
        sorted = self._sort_by_latency(hexels, hemi, list(function_hier.keys())) 

        hier = deepcopy(function_hier)               # we will modify function_hier so copy
        n_partitions = 1 / len(function_hier.keys()) # partition y-axis
        part_idx = 0                                 # pointer into the current partition out of [0, n_partitions). We go top-down cus of this.
                                                     # it ensures we order our nodes according to their level, with input at the bottom.
        graph = {}                                   # format: node_name : [magnitude, color, position, in_edges]
        while len(hier.keys()) > 0:
            top_level = self._get_top_level_functions(hier) # get function that is a child, i.e., it doesn't have any arrows going out of it
            for f in top_level:
                # We do the following:
                # if f == input_node:
                #     default_settings
                # else:
                #     create_spike_for_every_pairing
                #     for parent in parents:
                #          add spike from final parent node to the first f node. 
                # While doing this, if we encounter empty pairings for f or any parent, we skip them.
                
                if f in inputs:
                    # input node default size is 100.
                    hier.pop(f)
                    graph[f] = Node(100, (0, 1 - n_partitions * part_idx), [])
                
                else:
                    parents = hier[f]     
                    hier.pop(f)        # make sure to pop so we don't get the same top level fs every loop.

                    best_pairings = (
                            sorted[f].left_best_pairings if hemi == 'leftHemisphere' else
                                sorted[f].right_best_pairings
                        )
                    
                    if len(best_pairings) == 0:
                        # ignore functions with no spikes at all.
                        continue

                    for idx, pair in enumerate(best_pairings):
                        # add null edges to subsequent spikes in best_pairings.
                        latency, magnitude = pair 
                        graph[f + '-' + str(idx)] = Node(-10 * np.log10(magnitude),
                                                        (latency, 1 - n_partitions * part_idx),
                                                        [f + '-' + str(idx - 1)] if idx != 0 else [])
                    
                    part_idx += 1  # increment partition index

                    for parent in parents:
                        # add edges coming from parents to f.
                        if parent in inputs:
                            # if the parent is a input node, there is only one node.
                            graph[f + '-0'].in_edges.append(parent)
                        else:
                            parent_pairings = (
                                sorted[parent].left_best_pairings if hemi == 'leftHemisphere' else 
                                    sorted[parent].right_best_pairings
                                )
                            
                            if len(parent_pairings) == 0:
                                # ignore this function
                                continue

                            # add an edge from the final spike of parent to first spike of current function.
                            graph[f + '-0'].in_edges.append(parent + '-' + str(len(parent_pairings) - 1))
        return graph
    
    def _get_top_level_functions(self, hier: Dict[str, List[str]]) -> set:
        """
            Returns the top-level function. A top-level function is at the highest level
            of the function hierarchy. It can be found as the function that does not appear
            in any of the lists of parents. I.e., it is a function that that does not feed into
            any other functions; it only has functions feeding into it.

            Params
            ------
                hier : dictionary that contains the function hierarchy (including inputs)
            Returns
            -------
                a set containing the top-level functions.
        """
        funcs_leftover = list(hier.keys())
        parent_funcs = [f for parents in hier.values() for f in parents]
        return set(funcs_leftover).difference(set(parent_funcs))

    def _sort_by_latency(self, hexels: Dict[str, IPPMHexel], hemi: str, functions: List[str]) -> Dict[str, IPPMHexel]:
        """
            Sort pairings by latency in increasing order inplace.

            Params
            ------
                hexels contains all the functions and hexel objects to sort.
                hemi : leftHemisphere or rightHemisphere
            Returns
            -------
                sorted hexels.
        """
        for function in functions:
            if function not in hexels.keys():
                # function was not detected in the hexels
                continue
            
            if hemi == 'leftHemisphere':
                hexels[function].left_best_pairings.sort(key=lambda x: x[0])
            else:
                hexels[function].right_best_pairings.sort(key=lambda x: x[0])

        return hexels




from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional
from urllib import request

from kymata.datasets.data_root import data_root_path
from kymata.entities.expression import HexelExpressionSet, SensorExpressionSet
from kymata.io.file import PathType
from kymata.io.nkg import load_expression_set

_SAMPLE_DATA_DIR_NAME = "tutorial_nkg_data"


class SampleDataset(ABC):
    """
    Info required to retrieve a dataset stored locally.

    Names in `.filenames` refer to local files, which (if `remote_root` is specified) are paired with identically
    named remote files.
    """
    def __init__(self,
                 name: str,
                 filenames: list[str],
                 data_root: Optional[PathType],
                 remote_root: Optional[str],
                 download: bool):

        self.name: str = name
        self.filenames: list[str] = filenames
        self.data_root: Path = Path(data_root_path(data_root), _SAMPLE_DATA_DIR_NAME)
        self.remote_root: str = remote_root

        # Create the default location, if it's not being overridden
        if data_root is None:
            self.data_root.mkdir(exist_ok=True)

        if download:
            self.download()

    @property
    def path(self) -> Path:
        return Path(self.data_root, self.name)

    def download(self):
        print(f"Downloading dataset: {self.name}")
        if self.remote_root is None:
            raise ValueError("No remote root provided")
        self.path.mkdir(exist_ok=True)
        for filename in self.filenames:
            remote = self.remote_root + "/" + filename
            local = Path(self.path, filename)
            if local.exists():
                print(f"Local file already exists: {local}")
            else:
                print(f"Downloading {remote} to {local}")
                request.urlretrieve(remote, local)


class SampleExpressionDataset(SampleDataset, ABC):
    @abstractmethod
    def to_expressionset(self) -> HexelExpressionSet:
        raise NotImplementedError()


class KymataMirror2023Q3Dataset(SampleExpressionDataset):
    def __init__(self, data_root: Optional[PathType] = None, download: bool = True):
        name = "kymata_mirror_Q3_2023"
        super().__init__(
            name=name,
            filenames=[
                "kymata_mirror_Q3_2023_expression_endtable.nkg",
            ],
            data_root=data_root,
            remote_root="https://kymata.org/assets_kymata_toolbox_tutorial_data/gridsearch-result-data/",
            download=download,
        )

    def to_expressionset(self) -> HexelExpressionSet:
        es = load_expression_set(from_path_or_file=Path(self.path, self.filenames[0]))
        assert isinstance(es, HexelExpressionSet)
        return es


class TVLInsLoudnessOnlyDataset(SampleExpressionDataset):
    def __init__(self, data_root: Optional[PathType] = None, download: bool = True):
        name = "TVL_2020_ins_loudness_only"
        super().__init__(
            name=name,
            filenames=[
                "TVL_2020_ins_loudness_only.nkg",
            ],
            data_root=data_root,
            remote_root="https://kymata.org/assets_kymata_toolbox_tutorial_data/gridsearch-result-data/",
            download=download,
        )

    def to_expressionset(self) -> HexelExpressionSet:
        es = load_expression_set(from_path_or_file=Path(self.path, self.filenames[0]))
        assert isinstance(es, HexelExpressionSet)
        return es


class TVLDeltaInsTC1LoudnessOnlyDataset(SampleExpressionDataset):
    def __init__(self, data_root: Optional[PathType] = None, download: bool = True):
        name = "TVL_2020_delta_ins_tontop_chan1_loudness_only"
        super().__init__(
            name=name,
            filenames=[
                "TVL_2020_delta_ins_tontop_chan1_loudness_only.nkg"
            ],
            data_root=data_root,
            remote_root="https://kymata.org/assets_kymata_toolbox_tutorial_data/gridsearch-result-data/",
            download=download,
        )

    def to_expressionset(self) -> HexelExpressionSet:
        es = load_expression_set(from_path_or_file=Path(self.path, self.filenames[0]))
        assert isinstance(es, HexelExpressionSet)
        return es


class TVLDeltaInsTC1LoudnessOnlySensorsDataset(SampleExpressionDataset):
    def __init__(self, data_root: Optional[PathType] = None, download: bool = True):
        name = "TVL_2020_delta_ins_tontop_chan1_loudness_only_sensors"
        super().__init__(
            name=name,
            filenames=[
                "TVL_2020_delta_ins_tontop_chan1_loudness_only_sensors.nkg"
            ],
            data_root=data_root,
            remote_root="https://kymata.org/assets_kymata_toolbox_tutorial_data/gridsearch-result-data/",
            download=download,
        )

    def to_expressionset(self) -> SensorExpressionSet:
        es = load_expression_set(from_path_or_file=Path(self.path, self.filenames[0]))
        assert isinstance(es, SensorExpressionSet)
        return es


def delete_dataset(local_dataset: SampleDataset):
    from shutil import rmtree
    # Make sure it's not silent
    print(f"Deleting dataset {local_dataset.name}")
    # Only allow deletion if the specified url is within the data dir
    assert data_root_path() in local_dataset.path.parents, "Cannot delete dataset outside of data root directory"
    if not local_dataset.path.exists():
        # Nothing to delete
        print(f"{str(local_dataset.path)} doesn't exist")
        return

    # We passed the checks, do the delete
    rmtree(local_dataset.path)



from __future__ import annotations

import os
from dataclasses import dataclass
from itertools import cycle
from pathlib import Path
from statistics import NormalDist
from typing import Optional, Sequence, NamedTuple, Any, Type, Literal
from warnings import warn

import numpy as np
from matplotlib import pyplot
from matplotlib.colors import to_hex, LinearSegmentedColormap
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from matplotlib.ticker import FixedLocator
from mne import SourceEstimate
from numpy.typing import NDArray
from pandas import DataFrame
from seaborn import color_palette

from kymata.entities.expression import HexelExpressionSet, SensorExpressionSet, ExpressionSet, DIM_FUNCTION
from kymata.entities.functions import Function
from kymata.math.p_values import p_to_logp
from kymata.math.rounding import round_down, round_up
from kymata.plot.layouts import get_meg_sensor_xy, get_eeg_sensor_xy, get_meg_sensors, get_eeg_sensors

transparent = (0, 0, 0, 0)

# log scale: 10 ** -this will be the ytick interval and also the resolution to which the ylims will be rounded
_MAJOR_TICK_SIZE = 50


class _AxName:
    """Canonical names for the axes."""
    top    = "top"
    bottom = "bottom"
    main   = "main"
    minimap_top    = "minimap top"
    minimap_bottom = "minimap bottom"
    minimap_main   = "minimap main"


@dataclass
class _MosaicSpec:
    mosaic: list[list[str]]
    width_ratios: list[float] | None
    fig_size: tuple[float, float]
    expression_yaxis_label_xpos: float
    subplots_adjust_kwargs: dict[str, float] = None

    def __post_init__(self):
        if self.subplots_adjust_kwargs is None:
            self.subplots_adjust_kwargs = dict()

    def to_subplots(self) -> tuple[pyplot.Figure, dict[str, pyplot.Axes]]:
        return pyplot.subplot_mosaic(
            self.mosaic,
            width_ratios=self.width_ratios,
            figsize=self.fig_size)


def _minimap_mosaic(paired_axes: bool, show_minimap: bool, expression_set_type: Type[ExpressionSet], fig_size: tuple[float, float]) -> _MosaicSpec:
    # Set defaults:
    if show_minimap:
        width_ratios = [1, 3]
        subplots_adjust = {
            "hspace": 0, "wspace": 0.25,
            "left": 0.02, "right": 0.8,
        }
        # Place next to the expression plot yaxis
        yaxis_label_xpos = width_ratios[0]/(width_ratios[1]+width_ratios[0]) - 0.04
    else:
        width_ratios = None
        subplots_adjust = {
            "hspace": 0,
            "left": 0.08, "right": 0.84,
        }
        yaxis_label_xpos = 0.04

    if paired_axes:
        if show_minimap:
            if expression_set_type == HexelExpressionSet:
                spec = [
                    [_AxName.minimap_top,    _AxName.top],
                    [_AxName.minimap_bottom, _AxName.bottom],
                ]
            elif expression_set_type == SensorExpressionSet:
                spec = [
                    [_AxName.minimap_main,    _AxName.top],
                    [_AxName.minimap_main, _AxName.bottom],
                ]
            else:
                raise NotImplementedError()
        else:
            spec = [
                [_AxName.top],
                [_AxName.bottom],
            ]
    else:
        if show_minimap:
            spec = [
                [_AxName.minimap_main, _AxName.main],
            ]
        else:
            spec = [
                [_AxName.main],
            ]

    return _MosaicSpec(mosaic=spec, width_ratios=width_ratios, fig_size=fig_size,
                       subplots_adjust_kwargs=subplots_adjust,
                       expression_yaxis_label_xpos=yaxis_label_xpos)


def _hexel_minimap_data(expression_set: HexelExpressionSet, alpha_logp: float, show_functions: list[str]) -> tuple[NDArray, NDArray]:
    """
    Generates data arrays for a minimap visualization of significant hexels in a HexelExpressionSet.

    Args:
        expression_set (HexelExpressionSet): The set of hexel expressions to analyze.
        alpha_logp (float): The logarithm of the p-value threshold for significance. Hexels with values
            below this threshold are considered significant.
        show_functions (list[str]): A list of function names to consider for significance. Only these
            functions will be checked for significant hexels.

    Returns:
        tuple[NDArray, NDArray]: A tuple containing two arrays (one for the left hemisphere and one for
        the right hemisphere). Each array has a length equal to the number of hexels in the respective
        hemisphere, with entries:
            - 0, if no function is ever significant for this hexel.
            - i + 1, where i is the index of the function (from show_functions) that is significant
              for this hexel.

    Notes:
        This function identifies which hexels are significant for the given functions based on a provided
        significance threshold. It returns arrays for both the left and right hemispheres, where each entry
        indicates whether the hexel is significant for any function and, if so, which function it is
        significant for.
    """
    data_left = np.zeros((len(expression_set.hexels_left),))
    data_right = np.zeros((len(expression_set.hexels_right),))
    best_functions_left, best_functions_right = expression_set.best_functions()
    best_functions_left = best_functions_left[best_functions_left["value"] < alpha_logp]
    best_functions_right = best_functions_right[best_functions_right["value"] < alpha_logp]
    for function_i, function in enumerate(expression_set.functions,
                                          # 1-indexed, as 0 will refer to transparent
                                          start=1):
        if function not in show_functions:
            continue
        significant_hexel_names_left = best_functions_left[best_functions_left[DIM_FUNCTION] == function][expression_set.channel_coord_name]
        hexel_idxs_left = np.searchsorted(expression_set.hexels_left, significant_hexel_names_left.to_numpy())
        data_left[hexel_idxs_left] = function_i

        significant_hexel_names_right = best_functions_right[best_functions_right[DIM_FUNCTION] == function][expression_set.channel_coord_name]
        hexel_idxs_right = np.searchsorted(expression_set.hexels_right, significant_hexel_names_right.to_numpy())
        data_right[hexel_idxs_right] = function_i

    return data_left, data_right


def _plot_function_expression_on_axes(ax: pyplot.Axes, function_data: DataFrame, color, sidak_corrected_alpha: float, filled: bool):
    """
    Returns:
        x_min, x_max, y_min, y_max
            logp values for axis limits
            Note: *_min and *_max values are np.Inf and -np.Inf respectively if x or y is empty
                  (so they can be added to min() and max() without altering the result).
    """
    x = function_data["latency"].values * 1000  # Convert to milliseconds
    y = function_data["value"].values
    c = np.where(np.array(y) <= sidak_corrected_alpha, color, "black")
    ax.vlines(x=x, ymin=1, ymax=y, color=c)
    ax.scatter(x, y, facecolors=c if filled else 'none', s=20, edgecolors=c)

    x_min = x.min() if len(x) > 0 else np.Inf
    x_max = x.max() if len(x) > 0 else -np.Inf
    # Careful, the y value is inverted, with y==1 on the origin and y<1 away from the origin.
    # "y_min" here is real absolute min value in the data (closest to zero)
    y_min = y.min() if len(y) > 0 else np.Inf
    y_max = y.max() if len(y) > 0 else -np.Inf

    return x_min, x_max, y_min, y_max


class AxisAssignment(NamedTuple):
    axis_name: str
    axis_channels: list


sensor_left_right_assignment: tuple[AxisAssignment, AxisAssignment] = (
    AxisAssignment(axis_name="left",
                   axis_channels=[
                       sensor
                       for sensor, (x, y) in get_meg_sensor_xy().items()
                       if x <= 0
                   ] + [
                       sensor
                       for sensor, (x, y) in get_eeg_sensor_xy().items()
                       if x <= 0
                   ]),
    AxisAssignment(axis_name="right",
                   axis_channels=[
                       sensor
                       for sensor, (x, y) in get_meg_sensor_xy().items()
                       if x >= 0
                   ] + [
                       sensor
                       for sensor, (x, y) in get_eeg_sensor_xy().items()
                       if x >= 0
                   ]),
)


def _plot_minimap_sensor(expression_set: ExpressionSet, minimap_axis: pyplot.Axes, colors: dict[str, str], alpha_logp: float):
    raise NotImplementedError("Minimap not yet implemented for sensor data")


def _plot_minimap_hexel(expression_set: HexelExpressionSet,
                        show_functions: list[str],
                        lh_minimap_axis: pyplot.Axes, rh_minimap_axis: pyplot.Axes,
                        view: str, surface: str,
                        colors: dict[str, Any], alpha_logp: float):
    # Ensure we have the FSAverage dataset downloaded
    from kymata.datasets.fsaverage import FSAverageDataset
    fsaverage = FSAverageDataset(download=True)
    os.environ["SUBJECTS_DIR"] = str(fsaverage.path)

    # Functions which aren't being shown need to be padded into the colormap, for indexing purposes, but should show up
    # as transparent
    colors = colors.copy()
    for function in expression_set.functions:
        if function not in show_functions:
            colors[function] = transparent

    # segment at index 0 will map to transparent
    # segment at index i will map to function of index i-1
    colormap = LinearSegmentedColormap.from_list("custom",
                                                 # Insert transparent for index 0
                                                 colors=[transparent] + [colors[f] for f in expression_set.functions],
                                                 # +1 for the transparency
                                                 N=len(expression_set.functions)+1)
    data_left, data_right = _hexel_minimap_data(expression_set, alpha_logp=alpha_logp, show_functions=show_functions)
    stc = SourceEstimate(data=np.concatenate([data_left, data_right]),
                         vertices=[expression_set.hexels_left, expression_set.hexels_right],
                         tmin=0, tstep=1)
    warn("Plotting on the fsaverage brain. Ensure that hexel numbers match those of the fsaverage brain.")
    plot_kwargs = dict(
        subject='fsaverage',
        surface=surface,
        views=view,
        colormap=colormap,
        smoothing_steps=2,
        background="white",
        spacing="ico5",
        brain_kwargs={
            "offscreen": True,  # offscreen here appears to not actually do anything in mne
        },
        time_viewer=False,
        colorbar=False,
        transparent=False,
        clim=dict(
            kind="value",
            lims=[0, len(expression_set.functions)/2, len(expression_set.functions)])
    )
    # Plot left view
    lh_brain = stc.plot(hemi="lh", **plot_kwargs)
    lh_brain_fig = pyplot.gcf()
    lh_minimap_axis.imshow(lh_brain.screenshot())
    hide_axes(lh_minimap_axis)
    pyplot.close(lh_brain_fig)
    # Plot right view
    rh_brain = stc.plot(hemi="rh", **plot_kwargs)
    rh_brain_fig = pyplot.gcf()
    rh_minimap_axis.imshow(rh_brain.screenshot())
    hide_axes(rh_minimap_axis)
    pyplot.close(rh_brain_fig)


def hide_axes(axes: pyplot.Axes):
    """Hide all axes markings from a pyplot.Axes."""
    axes.get_xaxis().set_visible(False)
    axes.get_yaxis().set_visible(False)
    axes.axis("off")


def expression_plot(
        expression_set: ExpressionSet,
        show_only: Optional[str | Sequence[str]] = None,
        paired_axes: bool = True,
        # Statistical kwargs
        alpha: float = 1 - NormalDist(mu=0, sigma=1).cdf(5),  # 5-sigma
        # Style kwargs
        color: Optional[str | dict[str, str] | list[str]] = None,
        ylim: Optional[float] = None,
        xlims: Optional[tuple[float | None, float | None]] = None,
        hidden_functions_in_legend: bool = True,
        title: str = "Function expression",
        fig_size: tuple[float, float] = (12, 7),
        # Display options
        minimap: bool = False,
        minimap_view: str = "lateral",
        minimap_surface: str = "inflated",
        show_only_sensors: Optional[Literal["eeg", "meg"]] = None,
        # I/O args
        save_to: Optional[Path] = None,
        overwrite: bool = True,
        show_legend: bool = True,
        legend_display: dict[str, str] | None = None,
) -> pyplot.Figure:
    """
    Generates a plot of function expressions over time with optional display customizations.

    Args:
        expression_set (ExpressionSet): The set of expressions to plot, containing functions and associated data.
        show_only (Optional[str | Sequence[str]], optional): A string or a sequence of strings specifying which
            functions to plot.
            If None, all functions in the expression_set will be plotted. Default is None.
        paired_axes (bool, optional): When True, shows the expression plot split into left and right axes.
            When False, all points are shown on the same axis. Default is True.
        alpha (float, optional): Significance level for statistical tests, defaulting to a 5-sigma threshold.
        color (Optional[str | dict[str, str] | list[str]], optional): Color settings for the plot. Can be a single
            color, a dictionary mapping function names to colors, or a list of colors. Default is None.
        ylim (Optional[float], optional): The y-axis limit (p-value). Use log10 of the desired value — e.g. if the
            desired limit is 10^-100, supply ylim=-100. If None, it will be determined automatically. Default is None.
        xlims (tuple[Optional[float], Optional[float]], optional): The x-axis limits as a tuple (in ms). None to use
            default values, or set either entry to None to use the default for that value. Default is (-100, 800).
        hidden_functions_in_legend (bool, optional): If True, includes non-plotted functions in the legend.
            Default is True.
        title (str, optional): Title over the top axis in the figure. Default is "Function expression".
        fig_size (tuple[float, float], optional): Figure size in inches. Default is (12, 7).
        minimap (bool, optional): If True, displays a minimap of the expression data. Default is False.
        minimap_view (str, optional): The view type for the minimap, either "lateral" or other specified views.
            Valid options are:
            `"lateral"`: From the left or right side such that the lateral (outside) surface of the given hemisphere is
                         visible.
            `"medial"`: From the left or right side such that the medial (inside) surface of the given hemisphere is
                        visible (at least when in split or single-hemi mode).
            `"rostral"`: From the front.
            `"caudal"`: From the rear.
            `"dorsal"`: From above, with the front of the brain pointing up.
            `"ventral"`: From below, with the front of the brain pointing up.
            `"frontal"`: From the front and slightly lateral, with the brain slightly tilted forward (yielding a view
                         from slightly above).
            `"parietal"`: From the rear and slightly lateral, with the brain slightly tilted backward (yielding a view
                          from slightly above).
            `"axial"`: From above with the brain pointing up (same as 'dorsal').
            `"sagittal"`: From the right side.
            `"coronal"`: From the rear.
            Default is `lateral`.
        minimap_surface (str, optional): The surface type for the minimap, such as "inflated". Default is "inflated".
        show_only_sensors (str, optional): Show only one type of sensors. "meg" for MEG sensors, "eeg" for EEG sensors.
            None to show all sensors. Supplying this value with something other than a SensorExpressionSet causes will
            throw an exception. Default is None.
        save_to (Optional[Path], optional): Path to save the generated plot. If None, the plot is not saved.
            Default is None.
        overwrite (bool, optional): If True, overwrite the existing file if it exists. Default is True.
        show_legend (bool, optional): If True, displays the legend. Default is True.
        legend_display (dict[str, str] | None, optional): Allows grouping of multiple functions under the same legend
            item. Provide a dictionary mapping true function names to display names. None applies no grouping.
            Default is None.

    Returns:
        pyplot.Figure: The matplotlib figure object containing the generated plot.

    Raises:
        FileExistsError: If the file already exists at save_to and overwrite is set to False.

    Notes:
        The function plots the expression data with options to customize the appearance and statistical
        significance thresholds. It supports different data types (e.g., HexelExpressionSet, SensorExpressionSet)
        and can handle paired axes for left/right hemisphere data.
    """

    # Default arg values
    if show_only is None:
        # Plot all
        show_only = expression_set.functions
    elif isinstance(show_only, str):
        show_only = [show_only]
    not_shown = [f for f in expression_set.functions if f not in show_only]

    if color is None:
        color = dict()
    elif isinstance(color, str):
        # Single string specified: use all that colour
        color = {f: color for f in show_only}
    elif isinstance(color, list):
        # List specified, then pair up in order
        assert len(color) == len(show_only)
        color = {f: c for f, c in zip(show_only, color)}

    # Default colours
    cycol = cycle(color_palette("Set1"))
    for function in show_only:
        if function not in color:
            color[function] = to_hex(next(cycol))

    best_functions = expression_set.best_functions()

    if paired_axes:
        if isinstance(expression_set, HexelExpressionSet):
            axes_names = ("left hemisphere", "right hemisphere")
            assert isinstance(best_functions, tuple)
        elif isinstance(expression_set, SensorExpressionSet):
            axes_names = ("left", "right")
            # Same functions passed, filtering done at channel level
            best_functions = (best_functions, best_functions)
        else:
            raise NotImplementedError()
    else:
        if isinstance(expression_set, HexelExpressionSet):
            raise NotImplementedError("HexelExpressionSets have preset hemisphere assignments")
        elif isinstance(expression_set, SensorExpressionSet):
            axes_names = ("", )
            # Wrap into list
            best_functions = (best_functions, )
        else:
            raise NotImplementedError()

    if isinstance(expression_set, HexelExpressionSet):
        n_channels = len(expression_set.hexels_left) + len(expression_set.hexels_right)
    elif isinstance(expression_set, SensorExpressionSet):
        n_channels = len(expression_set.sensors)
    else:
        raise NotImplementedError()

    chosen_channels = _restrict_channels(expression_set, best_functions, show_only_sensors)

    sidak_corrected_alpha = 1 - (
        (1 - alpha)
        ** (1 / (2
                 * len(expression_set.latencies)
                 * n_channels
                 * len(show_only))))

    sidak_corrected_alpha = p_to_logp(sidak_corrected_alpha)

    def _custom_label(function_name):
        if legend_display is not None:
            if function_name in legend_display.keys():
                return legend_display[function_name]
        return function_name

    mosaic = _minimap_mosaic(paired_axes=paired_axes, show_minimap=minimap, expression_set_type=type(expression_set),
                             fig_size=fig_size)

    fig: pyplot.Figure
    axes: dict[str, pyplot.Axes]
    fig, axes = mosaic.to_subplots()

    expression_axes_list: list[pyplot.Axes]
    if paired_axes:
        expression_axes_list = [axes[_AxName.top], axes[_AxName.bottom]]  # For iterating over in a predictable order
    else:
        expression_axes_list = [axes[_AxName.main]]

    fig.subplots_adjust(**mosaic.subplots_adjust_kwargs)

    custom_handles = []
    custom_labels = []
    data_x_min, data_x_max = np.Inf, -np.Inf
    data_y_min             = np.Inf
    for function in show_only:

        custom_label = _custom_label(function)
        if custom_label not in custom_labels:
            custom_handles.extend([Line2D([], [], marker='.', color=color[function], linestyle='None')])
            custom_labels.append(custom_label)

        # We have a special case with paired sensor data, in that some sensors need to appear
        # on both sides of the midline.
        if paired_axes and isinstance(expression_set, SensorExpressionSet):
            assign_left_right_channels = sensor_left_right_assignment
            # Some points will be plotted on one axis, filled, some on both, empty
            top_chans = set(assign_left_right_channels[0].axis_channels) & chosen_channels
            bottom_chans = set(assign_left_right_channels[1].axis_channels) & chosen_channels
            # Symmetric difference
            both_chans = top_chans & bottom_chans
            top_chans -= both_chans
            bottom_chans -= both_chans
            for ax, best_funs_this_ax, chans_this_ax in zip(expression_axes_list, best_functions, (top_chans, bottom_chans)):
                # Plot filled
                x_min, x_max, y_min, _y_max, = _plot_function_expression_on_axes(
                    function_data=best_funs_this_ax[(best_funs_this_ax[DIM_FUNCTION] == function)
                                                    & (best_funs_this_ax[expression_set.channel_coord_name].isin(chans_this_ax))],
                    color=color[function],
                    ax=ax, sidak_corrected_alpha=sidak_corrected_alpha, filled=True)
                data_x_min = min(data_x_min, x_min)
                data_x_max = max(data_x_max, x_max)
                data_y_min = min(data_y_min, y_min)
                # Plot empty
                x_min, x_max, y_min, _y_max, = _plot_function_expression_on_axes(
                    function_data=best_funs_this_ax[(best_funs_this_ax[DIM_FUNCTION] == function)
                                                    & (best_funs_this_ax[expression_set.channel_coord_name].isin(both_chans))],
                    color=color[function],
                    ax=ax, sidak_corrected_alpha=sidak_corrected_alpha, filled=False)
                data_x_min = min(data_x_min, x_min)
                data_x_max = max(data_x_max, x_max)
                data_y_min = min(data_y_min, y_min)

        # With non-sensor data, or non-paired axes, we can treat these cases together
        else:
            # As normal, plot appropriate filled points in each axis
            for ax, best_funs_this_ax in zip(expression_axes_list, best_functions):
                x_min, x_max, y_min, _y_max, = _plot_function_expression_on_axes(
                    function_data=best_funs_this_ax[(best_funs_this_ax[DIM_FUNCTION] == function)
                                                    & (best_funs_this_ax[expression_set.channel_coord_name].isin(chosen_channels))],
                    color=color[function],
                    ax=ax, sidak_corrected_alpha=sidak_corrected_alpha, filled=True)
                data_x_min = min(data_x_min, x_min)
                data_x_max = max(data_x_max, x_max)
                data_y_min = min(data_y_min, y_min)

    # Format shared axis qualities
    for ax in expression_axes_list:
        xlims = _get_best_xlims(xlims, data_x_min, data_x_max)
        ylim = _get_best_ylim(ylim, data_y_min)
        ax.set_xlim(*xlims)
        ax.set_ylim((0, ylim))
        ax.axvline(x=0, color='k', linestyle='dotted')
        ax.axhline(y=sidak_corrected_alpha, color='k', linestyle='dotted')
        ax.text(-50, sidak_corrected_alpha, 'α*',
                bbox={'facecolor': 'white', 'edgecolor': 'none'}, verticalalignment='center')
        ax.text(600, sidak_corrected_alpha, 'α*',
                bbox={'facecolor': 'white', 'edgecolor': 'none'}, verticalalignment='center')
        ax.set_yticks(_get_yticks(ylim))
        # Format yaxis as p-values (e.g. 10^{-50} instead of -50)
        pval_labels = [
            f"$10^{{{int(t)}}}$" if t != 0 else "1"  # Instead of 10^0
            for t in ax.get_yticks()
        ]
        ax.set_yticklabels(pval_labels)

    # Plot minimap
    if minimap:
        if isinstance(expression_set, SensorExpressionSet):
            _plot_minimap_sensor(expression_set, minimap_axis=axes[_AxName.minimap_main],
                                 colors=color, alpha_logp=sidak_corrected_alpha)
        elif isinstance(expression_set, HexelExpressionSet):
            _plot_minimap_hexel(expression_set,
                                show_functions=show_only,
                                lh_minimap_axis=axes[_AxName.minimap_top],
                                rh_minimap_axis=axes[_AxName.minimap_bottom],
                                view=minimap_view, surface=minimap_surface,
                                colors=color, alpha_logp=sidak_corrected_alpha)
        else:
            raise NotImplementedError()

    # Format one-off axis qualities
    if paired_axes:
        top_ax = axes[_AxName.top]
        bottom_ax = axes[_AxName.bottom]
        top_ax.set_xticklabels([])
        bottom_ax.invert_yaxis()
    else:
        top_ax = bottom_ax = axes[_AxName.main]
    top_ax.set_title(title)
    bottom_ax.set_xlabel('Latency (ms) relative to onset of the environment')
    bottom_ax_xmin, bottom_ax_xmax = bottom_ax.get_xlim()
    bottom_ax.xaxis.set_major_locator(FixedLocator(_get_xticks((bottom_ax_xmin, bottom_ax_xmax))))
    if paired_axes:
        top_ax.text(s=axes_names[0],
                    x=bottom_ax_xmin + 20, y=ylim * 0.95,
                    style='italic', verticalalignment='center')
        bottom_ax.text(s=axes_names[1],
                       x=bottom_ax_xmin + 20, y=ylim * 0.95,
                       style='italic', verticalalignment='center')
    fig.text(x=mosaic.expression_yaxis_label_xpos, y=0.5,
             s='p-value (with α at 5-sigma, Šidák corrected)',
             ha="center", va="center", rotation="vertical")
    if bottom_ax_xmin <= 0 <= bottom_ax_xmax:
        bottom_ax.text(s='   onset of environment   ',
                       x=0, y=0 if paired_axes else ylim/2,
                       color='black', fontsize='x-small',
                       bbox={'facecolor': 'white',
                             'edgecolor': 'none',
                             'pad':2
                             },
                       verticalalignment='center',
                       horizontalalignment='center',
                       rotation='vertical')

    # Legend for plotted function
    if show_legend:
        split_legend_at_n_functions = 15
        legend_n_col = 2 if len(custom_handles) > split_legend_at_n_functions else 2
        if hidden_functions_in_legend and len(not_shown) > 0:

            if len(not_shown) > split_legend_at_n_functions:
                legend_n_col = 2
            # Plot dummy legend for other functions which are included in model selection but not plotted
            custom_labels_not_shown = []
            dummy_patches = []
            for hidden_function in not_shown:
                custom_label = _custom_label(hidden_function)
                if custom_label not in custom_labels_not_shown:
                    custom_labels_not_shown.append(custom_label)
                    dummy_patches.append(Patch(color=None, label=custom_label))
            leg = bottom_ax.legend(labels=custom_labels_not_shown, fontsize="x-small", alignment="left",
                                   title="Non-plotted functions",
                                   ncol=legend_n_col,
                                   bbox_to_anchor=(1.02, -0.02), loc="lower left",
                                   handles=dummy_patches)
            for lh in leg.legend_handles:
                lh.set_alpha(0)
        top_ax.legend(handles=custom_handles, labels=custom_labels, fontsize='x-small', alignment="left",
                      title="Plotted functions",
                      ncol=legend_n_col,
                      loc="upper left", bbox_to_anchor=(1.02, 1.02))

    if save_to is not None:
        pyplot.rcParams['savefig.dpi'] = 300
        save_to = Path(save_to)

        if overwrite or not save_to.exists():
            pyplot.savefig(Path(save_to), bbox_inches='tight')
        else:
            raise FileExistsError(save_to)

    return fig


def _restrict_channels(expression_set: ExpressionSet, best_functions: tuple[DataFrame, ...], show_only_sensors: str | None):
    """Restrict to specific sensor type if requested."""
    if show_only_sensors is not None:
        if isinstance(expression_set, SensorExpressionSet):
            if show_only_sensors == "meg":
                chosen_channels = get_meg_sensors()
            elif show_only_sensors == "eeg":
                chosen_channels = get_eeg_sensors()
            else:
                raise NotImplementedError()
        else:
            raise ValueError("`show_only_sensors` only valid with sensor data.")
    else:
        if isinstance(expression_set, SensorExpressionSet):
            # All sensors
            chosen_channels = {
                sensor
                for best_funs_each_ax in best_functions
                for sensor in best_funs_each_ax[expression_set.channel_coord_name]
            }
        elif isinstance(expression_set, HexelExpressionSet):
            # All hexels
            chosen_channels = {
                sensor
                for best_funs_each_ax in best_functions
                for sensor in best_funs_each_ax[expression_set.channel_coord_name]
            }
        else:
            raise NotImplementedError()
    return chosen_channels


def _get_best_xlims(xlims, data_x_min, data_x_max):
    default_xlims = (-200, 800)
    if xlims is None:
        xlims = (None, None)
    xmin, xmax = xlims
    if xmin is None:
        xmin = min(default_xlims[0], data_x_min)
    if xmax is None:
        xmax = max(default_xlims[1], data_x_max)
    xlims = (xmin, xmax)
    return xlims


def _get_best_ylim(ylim: float | None, data_y_min):
    if ylim is not None:
        return ylim
    default_y_min = -1 * _MAJOR_TICK_SIZE
    ylim = min(default_y_min, data_y_min)
    # Round to nearest major tick
    major_tick = np.floor(ylim / _MAJOR_TICK_SIZE) * _MAJOR_TICK_SIZE
    ylim = major_tick
    return ylim


def _get_xticks(xlims: tuple[float, float]):
    xmin, xmax = xlims
    # Round to the nearest 100
    step = 100
    xmin = round_up(xmin, step)
    xmax = round_down(xmax, step)
    return np.arange(xmin, xmax + 1, step)


def _get_yticks(ylim):
    n_major_ticks = int(ylim / _MAJOR_TICK_SIZE) * -1
    last_major_tick = -1 * n_major_ticks * _MAJOR_TICK_SIZE
    return np.linspace(start=0, stop=last_major_tick, num=n_major_ticks + 1)


def plot_top_five_channels_of_gridsearch(
        latencies: NDArray,
        corrs: NDArray,
        function: Function,
        n_samples_per_split: int,
        n_reps: int,
        n_splits: int,
        auto_corrs: NDArray,
        log_pvalues: any,
        # I/O args
        save_to: Optional[Path] = None,
        overwrite: bool = True,
):
    """
    Generates correlation and p-value plots showing the top five channels of the gridsearch.

    Args:
        latencies (NDArray[any]): Array of latency values (e.g., time points in milliseconds) for the x-axis of the plots.
        corrs (NDArray[any]): Correlation coefficients array with shape (n_channels, n_conditions, n_splits, n_time_steps).
        function (Function): The function object whose name attribute will be used in the plot title.
        n_samples_per_split (int): Number of samples per split used in the grid search.
        n_reps (int): Number of repetitions in the grid search.
        n_splits (int): Number of splits in the grid search.
        auto_corrs (NDArray[any]): Auto-correlation values array used for plotting the function auto-correlation.
        log_pvalues (any): Array of log-transformed p-values for each channel and time point.
        save_to (Optional[Path], optional): Path to save the generated plot. If None, the plot is not saved. Default is None.
        overwrite (bool, optional): If True, overwrite the existing file if it exists. Default is True.

    Raises:
        FileExistsError: If the file already exists at save_to and overwrite is set to False.

    Notes:
        The function generates two subplots:

        - The first subplot shows the correlation coefficients over latencies for the top five channels.
        - The second subplot shows the corresponding p-values for these channels.
    """

    figure, axis = pyplot.subplots(1, 2, figsize=(15, 7))
    figure.suptitle(f'{function.name}: Plotting corrs and pvalues for top five channels')

    corr_avrs = np.mean(corrs[:, 0], axis=-2) ** 2 # (n_chans, n_derangs, n_splits, t_steps) -> (n_chans, t_steps)
    maxs = np.max(corr_avrs, axis=1)
    n_amaxs = 5
    amaxs = np.argpartition(maxs, -n_amaxs)[-n_amaxs:]
    amax = np.argmax(corr_avrs) // (n_samples_per_split // 2)
    amaxs = [i for i in amaxs if i != amax]  # + [209]

    axis[0].plot(latencies, np.mean(corrs[amax, 0], axis=-2).T, 'r-', label=amax)
    axis[0].plot(latencies, np.mean(corrs[amaxs, 0], axis=-2).T, label=amaxs)
    std_null = np.mean(np.std(corrs[:, 1], axis=-2), axis=0).T * 3 / np.sqrt(n_reps * n_splits)  # 3 pop std.s
    std_real = np.std(corrs[amax, 0], axis=-2).T * 3 / np.sqrt(n_reps * n_splits)
    av_real = np.mean(corrs[amax, 0], axis=-2).T

    axis[0].fill_between(latencies, -std_null, std_null, alpha=0.5, color='grey')
    axis[0].fill_between(latencies, av_real - std_real, av_real + std_real, alpha=0.25, color='red')

    peak_lat_ind = np.argmax(corr_avrs) % (n_samples_per_split // 2)
    peak_lat = latencies[peak_lat_ind]
    peak_corr = np.mean(corrs[amax, 0], axis=-2)[peak_lat_ind]
    print(f'{function.name}: peak lat: {peak_lat:.1f},   peak corr: {peak_corr:.4f}   [sensor] ind: {amax},   -log(pval): {-log_pvalues[amax][peak_lat_ind]:.4f}')

    auto_corrs = np.mean(auto_corrs, axis=0)
    axis[0].plot(latencies, np.roll(auto_corrs, peak_lat_ind) * peak_corr / np.max(auto_corrs), 'k--',
                    label='func auto-corr')

    axis[0].axvline(0, color='k')
    axis[0].legend()
    axis[0].set_title("Corr coef.")
    axis[0].set_xlabel('latencies (ms)')
    axis[0].set_ylabel('Corr coef.')

    axis[1].plot(latencies, -log_pvalues[amax].T, 'r-', label=amax)
    axis[1].plot(latencies, -log_pvalues[amaxs].T, label=amaxs)
    axis[1].axvline(0, color='k')
    axis[1].legend()
    axis[1].set_title("p-values")
    axis[1].set_xlabel('latencies (ms)')
    axis[1].set_ylabel('p-values')

    if save_to is not None:
        pyplot.rcParams['savefig.dpi'] = 300
        save_to = Path(save_to, function.name + '_gridsearch_top_five_channels.png')

        if overwrite or not save_to.exists():
            pyplot.savefig(Path(save_to))
        else:
            raise FileExistsError(save_to)

    pyplot.clf()
    pyplot.close()


def legend_display_dict(functions: list[str], display_name) -> dict[str, str]:
    """
    Creates a dictionary for the `legend_display` parameter of `expression_plot()`.

    This function maps each function name in the provided list to a single display name,
    which can be used to group multiple functions under one legend item in the plot.

    Args:
        functions (list[str]): A list of function names to be grouped under the same display name.
        display_name (str): The display name to be used for all functions in the list.

    Returns:
        dict[str, str]: A dictionary mapping each function name to the provided display name.
    """
    return {function: display_name for function in functions}



from logging import getLogger
from pathlib import Path
from typing import Optional

import numpy as np
from numpy.typing import NDArray
from scipy.io import loadmat

from kymata.entities.functions import Function
from kymata.io.file import PathType


_logger = getLogger(__file__)


def load_function(function_path_without_suffix: PathType, func_name: str, replace_nans: Optional[bool] = None,
                  n_derivatives: int = 0, bruce_neurons: tuple = (0, 10)) -> Function:
    function_path_without_suffix = Path(function_path_without_suffix)
    func: NDArray
    if 'neurogram' in func_name:
        _logger.info('USING BRUCE MODEL')
        if function_path_without_suffix.with_suffix(".npz").exists():
            func_dict = np.load(function_path_without_suffix.with_suffix(".npz"))
            func = func_dict[func_name]
            func = np.mean(func[bruce_neurons[0]:bruce_neurons[1]], axis=0)
        else:
            mat = loadmat(function_path_without_suffix.with_suffix('.mat'))['neurogramResults']
            func = np.array(mat[func_name][0][0])
            # func = np.mean(func[bruce_neurons[0]:bruce_neurons[1]], axis=0)
            tt = np.array(mat['t_'+func_name[-2:]][0][0])
            n_chans = func.shape[0]
            new_mr_arr = np.zeros((n_chans, 400_000))
            new_ft_arr = np.zeros((n_chans, 400_000))
            if func_name == 'neurogram_mr':
                for i in range(n_chans):
                    new_mr_arr[i] = np.interp(np.linspace(0, 400, 400_000 + 1)[:-1], tt[0], func[i])
            elif func_name == 'neurogram_ft':
                func = np.cumsum(func * (tt[0, 1] - tt[0, 0]), axis=-1)
                for i in range(n_chans):
                    func_interp = np.interp(np.linspace(0, 400, 400_000 + 1), tt[0], func[i])
                    new_ft_arr[i] = np.diff(func_interp, axis=-1)

            func_dict = {'neurogram_mr': new_mr_arr,
                         'neurogram_ft': new_ft_arr}
            np.savez(str(function_path_without_suffix.with_suffix(".npz")), **func_dict)

            func = func_dict[func_name]
            func = np.mean(func[bruce_neurons[0]:bruce_neurons[1]], axis=0)

    else:
        if not function_path_without_suffix.with_suffix(".npz").exists():
            if function_path_without_suffix.with_suffix(".mat").exists():
                convert_stimulisig_on_disk_mat_to_npz(function_path_without_suffix)
            else:
                raise FileNotFoundError(f"{function_path_without_suffix}: neither .mat nor .npz found.")

        assert function_path_without_suffix.with_suffix(".npz").exists()

        func_dict = np.load(str(function_path_without_suffix.with_suffix(".npz")))
        func = np.array(func_dict[func_name], dtype=float)

    n_nan = np.count_nonzero(np.isnan(func))
    if n_nan > 0:
        nan_message = f"Function contained {n_nan:,} NaNs, ({100*n_nan/np.prod(func.shape):.2f}%)"
        if replace_nans is None:
            raise ValueError(f"{nan_message}. "
                             "Consider replacing small numbers of unavoidable NaNs with zeros or the function mean "
                             "value.")
        elif replace_nans == "zero":
            _logger.warning(f"{nan_message}. Replacing with zeros.")
            func[np.isnan(func)] = 0
        elif replace_nans == "mean":
            _logger.warning(f"{nan_message}. Replacing with the function mean.")
            func[np.isnan(func)] = np.nanmean(func)
        else:
            raise NotImplementedError()

    assert not np.isnan(func).any()

    for _ in range(n_derivatives):
        func = np.convolve(func, [-1, 1], 'same')  # derivative

    return Function(
        name=func_name,
        values=func.flatten().squeeze(),
        sample_rate=1000,
    )


def convert_stimulisig_on_disk_mat_to_npz(function_path_without_suffix):
    """
    Converts a (legacy) <stimulisig>.mat file to a (current) Numpy <stimulisig>.npz file on disk. Supply a path without
    a suffix, and the file will be created in the same place with the same name but with a different suffix.
    """
    mat = loadmat(str(function_path_without_suffix.with_suffix(".mat")))['stimulisig']
    _logger.info('Saving .mat as .npz ...')
    func_dict = {}
    for key in mat.dtype.names:
        if key == 'name':
            continue
        func_dict[key] = np.array(mat[key][0, 0], dtype=np.float16)
        func_dict[key].reshape((1, -1))  # Unwrap if it's a split matlab stimulisig
    np.savez(str(function_path_without_suffix.with_suffix(".npz")), **func_dict)



from typing import Dict, List, Tuple

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import numpy as np
from scipy.interpolate import splev

from .builder import Node


class IPPMPlotter(object):
    def draw(self, 
             graph: Dict[str, Node],
             colors: Dict[str, str],
             title: str,
             figheight: int=5,
             figwidth: int=10):
        """
        Generates an acyclic, directed graph using the graph held in graph. Edges are generated using BSplines.

        Args:
            graph (Dict[str, Node]): Dictionary with keys as node names and values as Hexel objects.
                Contains nodes as keys and magnitude, position, and incoming edges in the Hexel object.
            colors (Dict[str, str]): Dictionary with keys as node names and values as colors in hexadecimal.
                Contains the color for each function. The nodes and edges are colored accordingly.
            title (str): Title of the plot.
            figheight (int, optional): Height of the plot. Defaults to 5.
            figwidth (int, optional): Width of the plot. Defaults to 10.
        """
        # first lets aggregate all of the information.
        hexel_x = [_ for _ in range(len(graph.keys()))]                    # x coordinates for nodes e.g., (x, y) = (hexel_x[i], hexel_y[i])
        hexel_y = [_ for _ in range(len(graph.keys()))]                    # y coordinates for nodes
        node_colors = [_ for _ in range(len(graph.keys()))]                # color for nodes
        node_sizes = [_ for _ in range(len(graph.keys()))]                 # size of nodes
        edge_colors = []
        bsplines = []
        for i, node in enumerate(graph.keys()):
            for function, color in colors.items():
                # search for function color.
                if function in node:
                    node_colors[i] = color
                    break

            node_sizes[i] = graph[node].magnitude
            hexel_x[i] = graph[node].position[0]
            hexel_y[i] = graph[node].position[1]

            pairs = []
            for inc_edge in graph[node].in_edges:
                # save edge coordinates and color the edge the same color as the finishing node.
                start = graph[inc_edge].position
                end = graph[node].position
                pairs.append([(start[0], start[1]), (end[0], end[1])])
                edge_colors.append(node_colors[i])

            bsplines += self._make_bspline_paths(pairs)
        
        fig, ax = plt.subplots()
    
        for path, color in zip(bsplines, edge_colors):
            ax.plot(path[0], path[1], color=color, linewidth='3', zorder=-1)

        ax.scatter(x=hexel_x, y=hexel_y, c=node_colors, s=node_sizes, zorder=1)
        
        legend = []
        for f in colors.keys():
            legend.append(Line2D([0], [0], marker='o', color='w', label=f, markerfacecolor=colors[f], markersize=15))

        plt.legend(handles=legend, loc='upper left')
        plt.title(title)

        ax.set_ylim(min(hexel_y) - 0.1, max(hexel_y) + 0.1)
        ax.set_yticklabels([])
        ax.yaxis.set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.set_xlabel('Latency (ms)')
        
        fig.set_figheight(figheight)
        fig.set_figwidth(figwidth)
        
        #plt.show()

    def _make_bspline_paths(self, hexel_coordinate_pairs: List[List[Tuple[float, float]]]) -> List[List[np.array]]:
        """
        Given a list of hexel positions pairs, return a list of
        b-splines. First, find the control points, and second
        create the b-splines from these control points.

        Args:
            hexel_coordinate_pairs (List[List[Tuple[float, float]]]): Each list contains the x-axis values and y-axis values
                for the start and end of a BSpline, e.g., [(0, 1), (1, 0)].

        Returns:
            List[List[np.array]]: A list of a list of np arrays. Each list contains two np.arrays. The first np.array contains
                the x-axis points and the second one contains the y-axis points. Together, they define a BSpline. Thus, it is
                a list of BSplines.
        """
        bspline_path_array = []
        for pair in hexel_coordinate_pairs:
            start_X = pair[0][0]
            start_Y = pair[0][1]
            end_X = pair[1][0]
            end_Y = pair[1][1]
        
            if start_X + 35 > end_X and start_Y == end_Y:
                # the nodes are too close to use a bspline. Null edge.
                # add 2d np array where the first element is xs and second is ys
                xs = np.linspace(start_X, end_X, 100, endpoint=True)
                ys = np.array([start_Y] * 100)
                bspline_path_array.append([xs, ys])
            else:
                ctr_pts = self._make_bspline_ctr_points(pair)
                bspline_path_array.append(self._make_bspline_path(ctr_pts))

        return bspline_path_array

    def _make_bspline_ctr_points(self, start_and_end_node_coordinates: List[Tuple[float, float]]) -> np.array:
        """
        Given the position of a start hexel and an end hexel, create
        a set of 6 control points needed for a b-spline.

        The first one and last one is the position of a start hexel
        and an end hexel themselves, and the intermediate four are
        worked out using some simple rules.

        Args:
            start_and_end_node_coordinates (List[Tuple[float, float]]): List containing the start and end coordinates for one edge.
                First tuple is start, second is end. First element in tuple is x coord, second is y coord.

        Returns:
            np.array: A list of tuples of coordinates. Each coordinate pair represents a control point.
        """

        start_X = start_and_end_node_coordinates[0][0]
        start_Y = start_and_end_node_coordinates[0][1]
        end_X = start_and_end_node_coordinates[1][0]
        end_Y = start_and_end_node_coordinates[1][1]
        
        if end_X < start_X:
            # reverse BSpline
            start_X, end_X = end_X, start_X
            start_Y, end_Y = end_Y, start_Y

        bspline_ctr_points = []
        bspline_ctr_points.append((start_X,start_Y))

        # first 2
        bspline_ctr_points.append((start_X + 20, start_Y))
        bspline_ctr_points.append((start_X + 25, start_Y))

        # second 2
        bspline_ctr_points.append((start_X + 30, end_Y))
        bspline_ctr_points.append((start_X + 35, end_Y))

        bspline_ctr_points.append((end_X, end_Y))

        bspline_ctr_points = np.array(bspline_ctr_points)

        return bspline_ctr_points

    def _make_bspline_path(self, ctr_points: np.array) -> List[np.array]:
        """
        With an input of six control points, return an interpolated
        b-spline path which corresponds to a curved edge from one node to another.

        Args:
            ctr_points (np.array): 2d np.array containing the coordinates of the center points.

        Returns:
            List[np.array]: A list of np.arrays that represent one BSpline path. The first list is a list of x-axis coordinates
                the second is a list of y-axis coordinates.
        """
        x = ctr_points[:, 0]
        y = ctr_points[:, 1]

        length = len(x)
        t = np.linspace(0, 1, length - 2, endpoint=True)
        t = np.append([0, 0, 0], t)
        t = np.append(t, [1, 1, 1])

        tck = [t, [x, y], 3]
        u3 = np.linspace(0, 1, (max(length * 2, 70)), endpoint=True)
        bspline_path = splev(u3, tck)

        return bspline_path


from pathlib import Path
from typing import Optional

from mne.datasets import fetch_fsaverage

from kymata.datasets.sample import SampleDataset
from kymata.io.file import PathType


class FSAverageDataset(SampleDataset):
    def __init__(self,
                 data_root: Optional[PathType] = None,
                 download: bool = True,
                 ):
        from mne.datasets._fsaverage.base import FSAVERAGE_MANIFEST_PATH

        # Filenames are handled by mne:
        filenames = []
        with Path(FSAVERAGE_MANIFEST_PATH, "root.txt").open("r") as fsaverage_manifest:
            for line in fsaverage_manifest:
                filenames.append(line.strip())
        with Path(FSAVERAGE_MANIFEST_PATH, "bem.txt").open("r") as bem_manifest:
            for line in bem_manifest:
                filenames.append("fsaverage/" + line.strip())

        super().__init__(
            name="fsaverage",
            data_root=data_root,
            remote_root=None,  # Remote paths are handled by mne
            download=download,
            filenames=filenames,
            )

    def download(self):
        print(f"Downloading dataset: {self.name}")
        self.path.mkdir(exist_ok=True)
        _dir = fetch_fsaverage(subjects_dir=str(self.path), verbose=__debug__)



import mne
from mne.evoked import Evoked
from mne.minimum_norm.inverse import (
    INVERSE_METHODS, _check_ch_names, _check_or_prepare, _assemble_kernel, _check_ori,
    _check_reference
)
from mne.utils import (
    _check_option,
    _validate_type,
)


def pick_channels_inverse_operator(ch_names, inv):
    """Exposes `mne.minimum_norm.inverse._pick_channels_inverse_operator(ch_names, inv)`."""
    from mne.minimum_norm.inverse import _pick_channels_inverse_operator
    return _pick_channels_inverse_operator(ch_names, inv)


def premorph_inverse_operator(
    morph_path,
    evoked,
    inverse_operator_path,
    lambda2=1.0 / 9.0,
    method='dSPM',
    pick_ori=None,
    prepared=False,
    label=None,
    method_params=None,
    return_residual=False,
    use_cps=True,
):
    inverse_operator = mne.minimum_norm.read_inverse_operator(inverse_operator_path, verbose=False)
    _validate_type(evoked, Evoked, "evoked")
    _check_reference(evoked, inverse_operator["info"]["ch_names"])
    _check_option("method", method, INVERSE_METHODS)
    _check_ori(pick_ori, inverse_operator["source_ori"], inverse_operator["src"])
    #
    #   Set up the inverse according to the parameters
    #
    nave = evoked.nave

    _check_ch_names(inverse_operator, evoked.info)

    inv = _check_or_prepare(
        inverse_operator, nave, lambda2, method, method_params, prepared, copy="non-src"
    )

    K, noise_norm, vertno, source_nn = _assemble_kernel(
        inv, label, method, pick_ori, use_cps=use_cps
    )

    morph = mne.read_source_morph(morph_path)
    morph_csr = morph.morph_mat

    premorphed_inverse_operator = morph_csr.dot(K)

    return inv, premorphed_inverse_operator, morph.vertices_to



"""
Functions for dealing with iterables.
"""

from itertools import groupby
from typing import Sequence

from numpy import ndarray
from numpy.typing import NDArray


def all_equal(sequence: Sequence) -> bool:
    """
    Check if all elements in the sequence are the same.

    This function tests whether all elements in the provided sequence are equal. It handles special cases for
    sequences of length 0 or 1, and includes specific handling for numpy arrays due to their unique behavior
    with equality checks.

    Args:
        sequence (Sequence): A sequence of elements to be checked for equality. This can be any iterable
                             including lists, tuples, or numpy arrays.

    Returns:
        bool: True if all elements in the sequence are equal, False otherwise.

    Notes:
        - An empty sequence returns True as there are no elements to compare (universal quantification over
          an empty set is true).
        - A sequence with one element returns True as the single element is trivially equal to itself.
        - For numpy arrays, the function ensures element-wise comparison using the `.all()` method to handle
          array equality correctly.
    """

    if len(sequence) == 0:
        # universal quantification over empty set is always true
        return True
    elif len(sequence) == 1:
        # One item is equal to itself
        return True
    elif isinstance(sequence[0], ndarray):
        # numpy arrays deal with equality weirdly

        # Check first two items are equal, and equal to the rest
        first: NDArray = sequence[0]
        if not isinstance(sequence[1], ndarray):
            return False
        second: NDArray = sequence[1]
        try:
            # noinspection PyUnresolvedReferences
            return (first == second).all() and all_equal(sequence[1:])
        except AttributeError:
            return False
    else:
        g = groupby(sequence)
        return next(g, True) and not next(g, False)



from copy import deepcopy
from statistics import NormalDist
from typing import Dict, List, Tuple, Optional

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN, MeanShift
from sklearn.preprocessing import normalize

from .constants import TIMEPOINTS, NUMBER_OF_HEXELS
from .custom_clusterers import MaxPooler, AdaptiveMaxPooler, CustomGMM
from .data_tools import IPPMHexel


class DenoisingStrategy(object):
    """Superclass for unsupervised clustering algorithms. Strategies should conform to this interface."""

    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            **kwargs
    ):
        """
        :param hemi: Either "rightHemisphere" or "leftHemisphere". Indicates which hemisphere to cluster.
        :param should_normalise:
                Indicates whether the preprocessing step of scaling the data to be unit length should be done.
                Unnormalised data places greater precedence on the latency dimension.
        :param cluster_only_latency:
                Indicates whether we should discard the magnitude dimension from the clustering algorithm.
                We can term this "density-based clustering" as it focuses on the density of points in time only.
        :param should_max_pool:
                Indicates whether we want to enforce the constraint that each function should have 1 spike.
                Equivalently, we take the maximum over all clusters and discard all other points. Setting this to
                True optimises performance in the context of IPPMs.
        """

        self._clusterer = None
        self._hemi = hemi
        self._should_normalise = should_normalise
        self._should_cluster_only_latency = should_cluster_only_latency
        self._should_max_pool = should_max_pool
        self._threshold_for_significance = DenoisingStrategy._estimate_threshold_for_significance(normal_dist_threshold)

    @staticmethod
    def _estimate_threshold_for_significance(x: float) -> float:
        """
        bonferroni corrected = probability / number_of_trials.

        :param x: indicates the probability of P(X < x) for NormalDist(0, 1)
        :returns: a float indicating that threshold for significance.
        """

        def __calculate_bonferroni_corrected_alpha(one_minus_probability):
            return 1 - pow((1 - one_minus_probability), (1 / (2 * TIMEPOINTS * NUMBER_OF_HEXELS)))

        probability_X_gt_x = 1 - NormalDist(mu=0, sigma=1).cdf(x)
        threshold_for_significance = __calculate_bonferroni_corrected_alpha(probability_X_gt_x)
        return threshold_for_significance

    def denoise(self, hexels: Dict[str, IPPMHexel]) -> Dict[str, IPPMHexel]:
        """
            For a set of functions, cluster their IPPMHexel and retain the most significant spikes per cluster.

            TODO: Isn't max pooling after clustering equivalent to max pooling before? Only difference is that
                  anomalies are excluded. In that case, we could just MAX(all_spikes) to get optimal strategy.

            TODO: Figure out how to manage side-effects

            Algorithm
            ---------
            For each function in hemi,
                1. We create a dataframe containing only significant spikes.
                    1.1) [Optional] Perform any preprocessing.
                        - should_normalise: Scale each dimension to have a total length of 1.
                        - cluster_only_latency: Remove magnitude dimension.
                2. Cluster using the clustering method of the child class.
                3. For each cluster, we take the most significant point and discard the rest.
                    3.1) [Optional] Perform any postprocessing.
                        - should_max_pool: Take the most significant point over all of the clusters. I.e., only 1
                                                 spike per function.

            :param hexels:
                The key is the function name and the IPPMHexel contains information about the function. Specifically,
                it contains a list of (Latency (ms), Magnitude (^10-x)) for each hemisphere. We cluster over one
                hemisphere.
            :return: A dictionary of functions as keys and a IPPMHexel containing the clustered time-series.
        """
        hexels = deepcopy(
            hexels)  # When we copy the input, it is because we don't want to modify the original structure.
        # These are called side effects, and they can introduce obscure bugs.

        for func, df in self._map_hexels_to_df(hexels):
            if len(df) == 0:
                hexels[func] = self._update_pairings(hexels[func], [])
                continue

            preprocessed_df = self._preprocess(df)

            self._clusterer = self._clusterer.fit(preprocessed_df)
            # It is important you don't use the preprocessed_df to get the denoised because
            # we do not want to plot the preprocessed latencies and magnitudes.
            denoised_time_series = self._get_denoised_time_series(df)

            hexels[func] = self._postprocess(hexels[func], denoised_time_series)

        return hexels

    def _map_hexels_to_df(self, hexels: Dict[str, IPPMHexel]) -> pd.DataFrame:
        """
        A generator used to transform each pair of key, IPPMHexel to a DataFrame containing significance spikes only.

        :param hexels: The dictionary we want to iterate over and transform into DataFrames.
        :returns: a clean DataFrame with columns ['Latency', 'Mag'] for each function key.
        """
        for func, hexel in hexels.items():
            significant_spikes = self._filter_out_insignificant_spikes(
                hexel.right_best_pairings if self._hemi == 'rightHemisphere' else hexel.left_best_pairings,
            )
            df = pd.DataFrame(significant_spikes, columns=['Latency', 'Mag'])
            yield func, df

    def _filter_out_insignificant_spikes(self, spikes: List[Tuple[float, float]]) -> List[List[float]]:
        """
        For a list of spikes, remove all that are not statistically significant and save them to a DataFrame.

        :param spikes: spikes we want to filter based on their statistical significance.
        :param df: dataframe we will save the spikes to. Assumption: first column is Latency, second is Mag.
        :returns: DataFrame that is a subset of spikes with only significant spikes.
        """
        significant_datapoints = []
        for latency, spike in spikes:
            if spike <= self._threshold_for_significance:
                significant_datapoints.append([latency, spike])
        return significant_datapoints

    def _update_pairings(self, hexel: IPPMHexel, denoised: List[Tuple[float, float]]) -> IPPMHexel:
        """
        :param hexel: We want to update this hexel to store the denoised spikes, with max pooling if desired.
        :param denoised: We want to save these spikes into hexel, overwriting the previous ones.
        :returns: IPPM hexel with its state updated.
        """
        if self._hemi == 'rightHemisphere':
            hexel.right_best_pairings = denoised
        else:
            hexel.left_best_pairings = denoised
        return hexel

    def _preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Currently, this can normalise or remove the magnitude dimension.

        :param df: Dataframe we want to preprocess.
        :return: if we cluster_only_latency, we return a numpy array. Else, dataframe.
        """

        def __extract_and_wrap_latency_dim(df_with_latency_col):
            return np.reshape(df_with_latency_col['Latency'], (-1, 1))

        mags = list(df["Mag"])
        if self._should_cluster_only_latency:
            df = __extract_and_wrap_latency_dim(df)
        if self._should_normalise:
            # Assumption: we only have latency, mag columns.
            """
                Short text on normalisation for future reference.
                https://www.kaggle.com/code/residentmario/l1-norms-versus-l2-norms
            """
            normed_latency_and_mag = np.c_[normalize(df, axis=0), mags]
            df = pd.DataFrame(normed_latency_and_mag, columns=["Latency", "Mag"])
        return df

    def _get_denoised_time_series(self, df: pd.DataFrame) -> List[Tuple[float, float]]:
        """
        For a given dataframe, extract out the most significant spike for each cluster along with its associated latency
        and ignore anomalies (label == -1.)

        Assumes self._clusterer.fit has been called already. Also, assumes anomalies are tagged as "-1".

        :param df: Contains cols ["Latency", "Mag"] and IPPMHexel time-series.
        :returns: Each tuple contains ("Latency", "Mag") of the most significant points (excl. anomalies).
        """

        def __keep_most_significant_per_label(labelled_df):
            return labelled_df.loc[labelled_df.groupby('Label')['Mag'].idxmin()]

        def __filter_out_anomalies(labelled_df):
            return labelled_df[labelled_df['Label'] != -1]

        def __convert_df_to_list(most_significant_points_df):
            return list(zip(most_significant_points_df['Latency'], most_significant_points_df['Mag']))

        df = deepcopy(df)
        df["Label"] = self._clusterer.labels_
        most_significant_points = __keep_most_significant_per_label(df)
        most_significant_points = __filter_out_anomalies(most_significant_points)
        return __convert_df_to_list(most_significant_points)

    def _postprocess(self, hexel: IPPMHexel, denoised_time_series: List[Tuple[float, float]]) -> IPPMHexel:
        """
        To postprocess, overwrite the hexel data with most significant points and perform any postprocessing steps,
        such as max pooling.

        :param hexel: We will overwrite the data in this hexel with denoised data.
        :param denoised_time_series: most significant points per cluster excluding anomalies.
        :return: hexel with denoised time-series saved and, optionally, max pooled.
        """
        hexel = self._update_pairings(hexel, denoised_time_series)
        if self._should_max_pool:
            hexel = self._perform_max_pooling(hexel)
        return hexel

    def _perform_max_pooling(self, hexel: IPPMHexel) -> IPPMHexel:
        """
        Enforce the constraint that there is only 1 spike for a specific function. It is basically max(all_spikes).

        :param hexel: Contains the time-series we want to take the maximum over.
        :returns: same hexel but with only one spike.
        """
        # we take minimum because smaller is more significant.
        if hexel.left_best_pairings and self._hemi == 'leftHemisphere':
            hexel.left_best_pairings = [min(hexel.left_best_pairings, key=lambda x: x[1])]
        elif hexel.right_best_pairings and self._hemi == 'rightHemisphere':
            hexel.right_best_pairings = [min(hexel.right_best_pairings, key=lambda x: x[1])]
        return hexel


class MaxPoolingStrategy(DenoisingStrategy):
    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            bin_significance_threshold: int = 15,
            bin_size: int = 25
    ):
        super().__init__(hemi, should_normalise, should_cluster_only_latency, should_max_pool, normal_dist_threshold)
        self._clusterer = MaxPooler(bin_significance_threshold, bin_size)


class AdaptiveMaxPoolingStrategy(DenoisingStrategy):
    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            bin_significance_threshold: int = 5,
            base_bin_size: int = 10
    ):
        super().__init__(hemi, should_normalise, should_cluster_only_latency, should_max_pool, normal_dist_threshold)
        self._clusterer = AdaptiveMaxPooler(bin_significance_threshold, base_bin_size)


class GMMStrategy(DenoisingStrategy):
    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            number_of_clusters_upper_bound: int = 5,
            covariance_type: str = 'full',
            max_iter: int = 1000,
            n_init: int = 8,
            init_params: str = 'kmeans',
            random_state: Optional[int] = None,
            should_evaluate_using_AIC: bool = False
    ):
        super().__init__(hemi, should_normalise, should_cluster_only_latency, should_max_pool, normal_dist_threshold)
        self._clusterer = CustomGMM(number_of_clusters_upper_bound, covariance_type, max_iter, n_init, init_params,
                                    random_state, should_evaluate_using_AIC)


class DBSCANStrategy(DenoisingStrategy):
    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            eps: int = 10,
            min_samples: int = 2,
            metric: str = 'euclidean',
            algorithm: str = 'auto',
            leaf_size: int = 30,
            n_jobs: int = -1,
            metric_params: Optional[dict] = None
    ):
        super().__init__(hemi, should_normalise, should_cluster_only_latency, should_max_pool, normal_dist_threshold)
        self._clusterer = DBSCAN(eps=eps,
                                 min_samples=min_samples,
                                 metric=metric,
                                 metric_params=metric_params,
                                 algorithm=algorithm,
                                 leaf_size=leaf_size,
                                 n_jobs=n_jobs)


class MeanShiftStrategy(DenoisingStrategy):
    def __init__(
            self,
            hemi: str,
            should_normalise: bool = False,
            should_cluster_only_latency: bool = False,
            should_max_pool: bool = False,
            normal_dist_threshold: float = 5,
            cluster_all: bool = False,
            bandwidth: float = 30,
            seeds: Optional[int] = None,
            min_bin_freq: int = 2,
            n_jobs: int = -1
    ):
        super().__init__(hemi, should_normalise, should_cluster_only_latency, should_max_pool, normal_dist_threshold)
        self._clusterer = MeanShift(bandwidth=bandwidth,
                                    seeds=seeds,
                                    min_bin_freq=min_bin_freq,
                                    cluster_all=cluster_all,
                                    n_jobs=n_jobs)



import json
from itertools import cycle
from statistics import NormalDist
from typing import Tuple, Dict, List
from collections import namedtuple

import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.preprocessing import normalize
from sklearn.metrics.pairwise import euclidean_distances
import requests
import seaborn as sns
from matplotlib.lines import Line2D
import math
from kymata.entities.expression import HexelExpressionSet, DIM_FUNCTION, DIM_LATENCY

Node = namedtuple('Node', 'magnitude position inc_edges')


class IPPMHexel(object):
    """
        Container to hold data about a hexel spike.
        
        Attributes
        ----------   
            function : the name of the function who caused the spike
            right_best_pairings : right hemisphere best pairings. pvalues are taken to the base 10 by default. latency is in milliseconds
            left_best_pairings : right hemisphere best pairings. same info as right for (latency, pvalue)
            description : optional written description
            github_commit : github commit of the function
    """
    
    def __init__(
                self, 
                function_name: str, 
                description: str=None, 
                github_commit: str=None,
            ): 
            self.function = function_name
            self.right_best_pairings = []
            self.left_best_pairings = []
            self.description = description
            self.github_commit = github_commit
            self.color = None

            self.input_stream = None
            
    def add_pairing(self, hemi: str, pairing: Tuple[float, float]):
        """
            Use this to add new pairings. Pair = (latency (ms), pvalue (log_10))

            Params
            ------
                hemi : leftHemisphere or rightHemisphere
                pairing : Corresponds to the best match to a hexel spike of form (latency (ms), pvalue (log_10))
        """
        if hemi == 'leftHemisphere':
            self.left_best_pairings.append(pairing)
        else:
            self.right_best_pairings.append(pairing)


def fetch_data(api: str) -> Dict[str, IPPMHexel]:
    """
        Fetches data from Kymata API and converts it into a dictionary of function names as keys
        and hexel objects as values. Advantage of dict is O(1) look-up and hexel object is readable
        access to attributes.
        
        Params
        ------
            api : URL of the API from which to fetch data
                
        Returns
        -------
            Dictionary containing data in the format [function name, hexel]
    """
    response = requests.get(api)
    resp_dict = json.loads(response.text)
    return build_hexel_dict_from_api_response(resp_dict)


def build_hexel_dict_from_expression_set(expression_set: HexelExpressionSet) -> Dict[str, IPPMHexel]:
    """
        Builds the dictionary from an ExpressionSet. This function builds a new dictionary
        which has function names (fast look-up) and only necessary data.

        Params
        ------
            dict_ : JSON dictionary of HTTP GET response object.

        Returns
        -------
            Dict of the format [function name, Hexel(func_name, id, left_pairings, right_pairings)]
    """
    best_functions_left, best_functions_right = expression_set.best_functions()
    hexels = {}
    for hemi in ['leftHemisphere', 'rightHemisphere']:
        best_functions = best_functions_left if hemi == "leftHemisphere" else best_functions_right
        for _idx, row in best_functions.iterrows():
            func = row[DIM_FUNCTION]
            latency = row[DIM_LATENCY] * 1000  # convert to ms
            pval = row["value"]
            if func not in hexels:
                hexels[func] = IPPMHexel(func)
            hexels[func].add_pairing(hemi, (latency, pval))
    return hexels


def build_hexel_dict_from_api_response(dict_: Dict) -> Dict[str, IPPMHexel]:
    """
        Builds the dictionary from response dictionary. Response dictionary has unneccesary 
        keys and does not have function names as keys. This function builds a new dictionary
        which has function names (fast look-up) and only necessary data.

        Params
        ------
            dict_ : JSON dictionary of HTTP GET response object.

        Returns
        -------
            Dict of the format [function name, Hexel(func_name, id, left_pairings, right_pairings)]
    """
    hexels = {}        
    for hemi in ['leftHemisphere', 'rightHemisphere']:
        for (_, latency, pval, func) in dict_[hemi]:
            # we have id, latency (ms), pvalue (log_10), function name.
            # discard id as it conveys no useful information
            if func not in hexels:
                # first time seeing function, so create key and hexel object.
                hexels[func] = IPPMHexel(func)
            
            hexels[func].add_pairing(hemi, (latency, pow(10, pval)))
    
    return hexels


def stem_plot(
        hexels: Dict[str, IPPMHexel],
        title: str,
        timepoints: int=201,
        y_limit: float=pow(10, -100),
        number_of_hexels: int=200000,
        figheight: int=7,
        figwidth: int=12,
        ):
    """
        Plots a stem plot using hexels.

        Params
        ------
            hexels : Contains function spikes in the form of a Hexel object. All pairings are found there.
            title : Title of plot.
    """
    # estimate significance parameter
    alpha = 1 - NormalDist(mu=0, sigma=1).cdf(5)      # 5-sigma
    bonferroni_corrected_alpha = 1-(pow((1-alpha),(1/(2*timepoints*number_of_hexels))))

    # assign unique color to each function
    cycol = cycle(sns.color_palette("hls", len(hexels.keys())))
    for _, hexel in hexels.items():
        hexel.color = matplotlib.colors.to_hex(next(cycol))

    fig, (left_hem_expression_plot, right_hem_expression_plot) = plt.subplots(nrows=2, ncols=1, figsize=(figwidth, figheight))
    fig.subplots_adjust(hspace=0)
    fig.subplots_adjust(right=0.84, left=0.08)

    custom_handles = []
    custom_labels = []
    for key, my_function in hexels.items():

        color = my_function.color
        label = my_function.function

        custom_handles.extend([Line2D([], [], marker='.', color=color, linestyle='None')])
        custom_labels.append(label)

        # left
        left = list(zip(*(my_function.left_best_pairings)))
        if len(left) != 0:
            x_left, y_left = left[0], left[1]
            left_color = np.where(np.array(y_left) <= bonferroni_corrected_alpha, color, 'black') # set all insignificant spikes to black
            left_hem_expression_plot.vlines(x=x_left, ymin=1, ymax=y_left, color=left_color)
            left_hem_expression_plot.scatter(x_left, y_left, color=left_color, s=20)

        # right
        right = list(zip(*(my_function.right_best_pairings)))
        if len(right) != 0:
            x_right, y_right = right[0], right[1]
            right_color = np.where(np.array(y_right) <= bonferroni_corrected_alpha, color, 'black') # set all insignificant spikes to black
            right_hem_expression_plot.vlines(x=x_right, ymin=1, ymax=y_right, color=right_color)
            right_hem_expression_plot.scatter(x_right, y_right, color=right_color, s=20)

    for plot in [right_hem_expression_plot, left_hem_expression_plot]:
        plot.set_yscale('log')
        plot.set_xlim(-200, 800)
        plot.set_ylim(1, y_limit)
        plot.axvline(x=0, color='k', linestyle='dotted')
        plot.axhline(y=bonferroni_corrected_alpha, color='k', linestyle='dotted')
        plot.text(-100, bonferroni_corrected_alpha, 'α*', bbox={'facecolor': 'white', 'edgecolor': 'none'}, verticalalignment='center')
        plot.text(600, bonferroni_corrected_alpha, 'α*', bbox={'facecolor': 'white', 'edgecolor': 'none'}, verticalalignment='center')
        plot.set_yticks([1, pow(10,-50), pow(10,-100)])

    left_hem_expression_plot.set_title(title)
    left_hem_expression_plot.set_xticklabels([])
    right_hem_expression_plot.set_xlabel('Latency (ms) relative to onset of the environment')
    right_hem_expression_plot.xaxis.set_ticks(np.arange(-200, 800+1, 100))
    right_hem_expression_plot.invert_yaxis()
    left_hem_expression_plot.text(-180, y_limit * 10000000, 'left hemisphere', style='italic', verticalalignment='center')
    right_hem_expression_plot.text(-180, y_limit * 10000000, 'right hemisphere', style='italic', verticalalignment='center')
    y_axis_label = 'p-value (with α at 5-sigma, Bonferroni corrected)'
    left_hem_expression_plot.text(-275, 1, y_axis_label, verticalalignment='center',rotation='vertical')
    right_hem_expression_plot.text(0, 1, '   onset of environment   ', color='white', fontsize='x-small', bbox={'facecolor': 'grey', 'edgecolor': 'none'}, verticalalignment='center', horizontalalignment='center', rotation='vertical')
    left_hem_expression_plot.legend(handles=custom_handles, labels=custom_labels, fontsize='x-small', bbox_to_anchor=(1.2, 1))

    plt.show()


def causality_violation_score(denoised_hexels: Dict[str, IPPMHexel], hierarchy: Dict[str, List[str]], hemi: str, inputs: List[str]) -> Tuple[float, int, int]:
    """
        Assumption: hexels are denoised. Otherwise, it doesn't really make sense to check the min/max latency of noisy hexels.

        A score calculated on denoised hexels that calculates the proportion of arrows in IPPM that are going backward in time.
        It assumes that the function hierarchy is correct, which may not always be correct, so you must use it with caution. 

        Algorithm
        ----------
        violations = 0
        total_arrows = 0
        for each func_name, parents_list in hierarchy:
            child_lat = min(hexels[func])
            for parent in parents_list:
                parent_lat = max(hexels[parent])
                if child_lat < parent_lat:
                    violations++
                total_arrows++
        return violations / total_arrows if total_arrows > 0 else 0
    """
    
    assert(hemi == 'rightHemisphere'  or hemi == 'leftHemisphere')

    def get_latency(func_hexels: IPPMHexel, mini: bool):
        return (min(func_hexels.left_best_pairings, key=lambda x: x[0]) if hemi == 'leftHemisphere' else
                min(func_hexels.right_best_pairings, key=lambda x: x[0])) if mini else (
            max(func_hexels.left_best_pairings, key=lambda x: x[0]) if hemi == 'leftHemisphere' else
            max(func_hexels.right_best_pairings, key=lambda x: x[0]))

    causality_violations = 0
    total_arrows = 0
    for func, inc_edges in hierarchy.items():
        # essentially: if max(parent_spikes_latency) > min(child_spikes_latency), there will be a backwards arrow in time.
        # arrows go from latest inc_edge spike to the earliest func spike

        if func in inputs:
            continue

        if hemi == 'leftHemisphere':
            if len(denoised_hexels[func].left_best_pairings) == 0:
                continue
        else:
            if len(denoised_hexels[func].right_best_pairings) == 0:
                continue
        
        child_latency = get_latency(denoised_hexels[func], mini=True)[0]
        for inc_edge in inc_edges:
            if inc_edge in inputs:
                # input node, so parent latency is 0
                parent_latency = 0
                if child_latency < parent_latency:
                    causality_violations += 1
                total_arrows += 1
                continue

            # We need to ensure the function has significant spikes
            if hemi == 'leftHemisphere':
                if len(denoised_hexels[inc_edge].left_best_pairings) == 0:
                    continue
            else:
                if len(denoised_hexels[inc_edge].right_best_pairings) == 0:
                    continue
                    
            parent_latency = get_latency(denoised_hexels[inc_edge], mini=False)[0]
            if child_latency < parent_latency:
                causality_violations += 1
            total_arrows += 1

    return (
        causality_violations / total_arrows if total_arrows != 0 else 0,
        causality_violations,
        total_arrows)

def function_recall(noisy_hexels: Dict[str, IPPMHexel], funcs: List[str], ippm_dict: Dict[str, Node], hemi: str) -> Tuple[float]:
    """
        This is the second scoring metric: function recall. It illustrates what proportion out of functions in the noisy hexels are detected as part of IPPM. E.g., 9 functions but only 8 found => 8/9 = function recall. Use this along with causality violation to evaluate IPPMs and analyse their strengths and weaknesses. 
        
        One thing to note is that the recall depends upon the nature of the dataset. If certain functions have no significant spikes, there is an inherent bias present in the dataset. We can never get the function recall to be perfect no matter what algorithm we employ. Therefore, the function recall is based on what we can actually do with a dataset. E.g., 9 functions in the hierarchy but in the noisy hexels we find only 7 of the 9 functions. Moreover, after denoising we find that there are only 6 functions in the hierarchy. The recall will be 6/7 rather than 6/9 since there were only 7 to be found to begin with.

        Params
        ------
        hexels: the noisy hexels that we denoise and feed into IPPMBuilder. It must be the same dataset.
        funcs: list of functions that are in our hierarchy. Don't include the input function, e.g., input_cochlear.
        ippm_dict: the return value from IPPMBuilder. It contains node names as keys and Node objects as values.
        hemi: leftHemisphere or rightHemisphere

        Returns
        -------
        A ratio indicating how many channels were incorporated into the IPPM out of all relevant channels.
    """
    assert(hemi == 'rightHemisphere' or hemi == 'leftHemisphere')

    # Step 1: Calculate significance level
    alpha = 1 - NormalDist(mu=0, sigma=1).cdf(5)      
    bonferroni_corrected_alpha = 1-(pow((1-alpha),(1/(2*201*200000))))
    funcs_present_in_data = 0
    detected_funcs = 0
    for func in funcs:
        pairings = noisy_hexels[func].right_best_pairings if hemi == 'rightHemisphere' else noisy_hexels[func].left_best_pairings
        for latency, spike in pairings:
            # Step 2: Find a pairing that is significant
            if spike <= bonferroni_corrected_alpha:
                funcs_present_in_data += 1
                
                # Step 3: Found a function, look in ippm_dict.keys() for the function.
                for node_name in ippm_dict.keys():
                    if func in node_name:
                        # Step 4: If found, then increment detected_funcs. Also increment funcs_pressent
                        detected_funcs += 1
                        break
                break

    # Step 3: Return [ratio, numerator, denominator] primarily because both the denominator and numerator can vary.
    return (detected_funcs / funcs_present_in_data if funcs_present_in_data > 0 else 0, 
            detected_funcs, 
            funcs_present_in_data)
    

def convert_to_power10(hexels: Dict[str, IPPMHexel]) -> Dict[str, IPPMHexel]:
    """
        Utility function to take data from the .nkg format and convert it to power of 10, so it can be used for IPPMs.

        Parameters
        ------------
        hexels: dict function_name as key and hexel object as value. Hexels contain pairings for left/right.
        
        Returns
        --------
        same dict but the pairings are all raised to power x. E.g., pairings = [(lat1, x), ..., (latn, xn)] -> [(lat1, 10^x), ..., (latn, 10^xn)]
    """
    for func, hexel in hexels.items():
        hexels[func].right_best_pairings = list(map(lambda x: (x[0], math.pow(10, x[1])), hexels[func].right_best_pairings))
        hexels[func].left_best_pairings = list(map(lambda x: (x[0], math.pow(10, x[1])), hexels[func].left_best_pairings))
    return hexels


def remove_excess_funcs(to_retain: List[str], hexels: Dict[str, IPPMHexel]) -> Dict[str, IPPMHexel]:
    """
        Utility function to distill the hexels down to a subset of functions. Use this to visualise a subset of functions for time-series.
        E.g., you want the time-series for one function, so just pass it wrapped in a list as to_retain

        Parameters
        ----------
        to_retain: list of functions we want to retain in the hexels dict
        hexels: hexels: dict function_name as key and hexel object as value. Hexels contain pairings for left/right.

        Returns
        -------
        hexels but all functions that aren't in to_retain are filtered.
    """

    funcs = list(hexels.keys()) # need this because we can't remove from the dict while also iterating over it.
    for func in funcs:
        if func not in to_retain:
            # delete
            hexels.pop(func)
    return hexels


def plot_k_dist_1D(pairings: List[Tuple[float, float]], k: int=4, normalise: bool=False):
    """
        This could be optimised further but since we aren't using it, we can leave it as it is.
            
        A utility function to plot the k-dist graph for a set of pairings. Essentially, the k dist graph plots the distance
        to the kth neighbour for each point. By inspecting the gradient of the graph, we can gain some intuition behind the density of 
        points within the dataset, which can feed into selecting the optimal DBSCAN hyperparameters.

        For more details refer to section 4.2 in https://www.dbs.ifi.lmu.de/Publikationen/Papers/KDD-96.final.frame.pdf

        Parameters
        ----------
        pairings: list of pairings extracted from a hexel. It contains the pairings for one function and one hemisphere
        k: the k we use to find the kth neighbour. Paper above advises to use k=4.
        normalise: whether to normalise before plotting the k-dist. It is important because the k-dist then equally weights both dimensions.

        Returns
        -------
        Nothing but plots a graph.
    """
    
    alpha = 3.55e-15
    X = pd.DataFrame(columns=['Latency'])
    for latency, spike in pairings:
        if spike <= alpha:
            X.loc[len(X)] = [latency]

    if normalise:
        X = normalize(X)
            
    distance_M = euclidean_distances(X) # rows are points, columns are other points same order with values as distances
    k_dists = []
    for r in range(len(distance_M)):
        sorted_dists = sorted(distance_M[r], reverse=True) # descending order
        k_dists.append(sorted_dists[k]) # store k-dist
    sorted_k_dists = sorted(k_dists, reverse=True)
    plt.plot(list(range(0, len(sorted_k_dists))), sorted_k_dists)
    plt.show()


def copy_hemisphere(
    hexels_to: Dict[str, IPPMHexel], 
    hexels_from: Dict[str, IPPMHexel], 
    hemi_to: str,
    hemi_from: str,
    func: str = None):
    """
        Utility function to copy a hemisphere onto another one. The primary use-case is to plot the denoised hemisphere against the
        noisy hemisphere using the same hexel object. I.e., copy right hemisphere to left; denoise on right; plot right vs left.

        Parameters
        ----------
        hexels_to: Hexels we are writing into. Could be (de)noisy hexels.
        hexels_from: Hexels we are copying from
        hemi_to: the hemisphere we index into when we write into hexels_to. E.g., hexels_to[hemi_to] = hexels_from[hemi_from]
        hemi_from: the hemisphere we index into when we copy the hexels from hexels_from. 
        func: if func != None, we only copy one function. Otherwise, we copy all.
        
        Returns
        -------
        Nothing, everything is done in-place. I.e., hexels_to is now updated.
    """
    if func:
        # copy only one function
        if hemi_to == 'rightHemisphere' and hemi_from == 'rightHemisphere':
            hexels_to[func].right_best_pairings = hexels_from[func].right_best_pairings
        elif hemi_to == 'rightHemisphere' and hemi_from == 'leftHemisphere':
            hexels_to[func].right_best_pairings = hexels_from[func].left_best_pairings
        elif hemi_to == 'leftHemisphere' and hemi_from == 'rightHemisphere':
            hexels_to[func].left_best_pairings = hexels_from[func].right_best_pairings
        else:
            hexels_to[func].left_best_pairings = hexels_from[func].left_best_pairings
        return
        
    for func, hexel in hexels_from.items():
        if hemi_to == 'rightHemisphere' and hemi_from == 'rightHemisphere':
            hexels_to[func].right_best_pairings = hexels_from[func].right_best_pairings
        elif hemi_to == 'rightHemisphere' and hemi_from == 'leftHemisphere':
            hexels_to[func].right_best_pairings = hexels_from[func].left_best_pairings
        elif hemi_to == 'leftHemisphere' and hemi_from == 'rightHemisphere':
            hexels_to[func].left_best_pairings = hexels_from[func].right_best_pairings
        else:
            hexels_to[func].left_best_pairings = hexels_from[func].left_best_pairings


def plot_denoised_vs_noisy(hexels: Dict[str, IPPMHexel], clusterer, title: str):
    """
        Utility function to plot the noisy and denoised versions. It runs the supplied clusterer and then copies the denoised hexels, which
        are fed into a stem plot.

        Parameters
        ----------
        hexels: hexels we want to denoise then plot
        clusterer: A child class of DenoisingStrategy that implements .cluster
        title: title of plot

        Returns
        -------
        Nothing but plots a graph.
    """
    denoised_hexels = clusterer.cluster(hexels, 'rightHemisphere')
    copy_hemisphere(denoised_hexels, hexels, 'leftHemisphere', 'rightHemisphere')
    stem_plot(denoised_hexels, title)



from typing import Dict, List

from kymata.ippm.ippm_builder import IPPMBuilder
from kymata.ippm.data_tools import IPPMHexel
from kymata.ippm.ippm_plotter import IPPMPlotter


class IPPM:
    def __init__(self,
                 hexels: Dict[str, IPPMHexel],
                 inputs: List[str],
                 hierarchy: Dict[str, List[str]],
                 hemisphere: str):
        self._builder = IPPMBuilder(hexels, inputs, hierarchy, hemisphere)
        self._plotter = IPPMPlotter()

    def create_ippm(self, colors: Dict[str, str]):
        graph = self._builder.build_graph_dict()
        self._plotter.plot(graph, colors)
